import os

from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
import pytest

import numpy as np
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import ConvexBoundaryComp,\
    CircleBoundaryConstraint, CircleBoundaryComp
from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
from topfarm.plotting import NoPlot, PlotComp
from topfarm.tests import npt


def testCircle():
    b = CircleBoundaryComp(3, [1, 2], 3)
    np.testing.assert_array_almost_equal(((b.xy_boundary - [1, 2])**2).sum(1), 3**2)


def test_TopFarmProblem_with_cirleboundary_penalty():
    optimal = np.array([(0, 0)])
    desvar = dict(zip('xy', optimal.T))
    b = CircleBoundaryConstraint([1, 2], 3)
    driver = SimpleGADriver()
    tf = TopFarmProblem(
        desvar,
        DummyCost(optimal, 'xy'),
        constraints=[b],
        driver=driver)
    tf.evaluate()
    tf.plot_comp.show()
    np.testing.assert_array_almost_equal(
        ((b.constraintComponent.xy_boundary - [1, 2])**2).sum(1), 3**2)
    npt.assert_array_less(tf.evaluate({'x': [3.9], 'y': [2]})[0], 1e10)
    npt.assert_array_less(1e10, tf.evaluate({'x': [4.1], 'y': [2]})[0])


def test_TopFarmProblem_with_cirleboundary_constraint():
    optimal = np.array([(0, 0)])
    desvar = dict(zip('xy', optimal.T))

    b = CircleBoundaryConstraint([2, 2], 2)
    tf = TopFarmProblem(
        desvar,
        DummyCost(optimal, 'xy'),
        constraints=[b])
    _, state, _ = tf.optimize()

    npt.assert_array_less((state['x'] - 2)**2 + (state['y'] - 2)**2, 9)


def test_TopFarmProblem_with_cirleboundary_constraint_and_limits():
    optimal = np.array([(0, 0)])
    desvar = {'x': ([0], 1, 4), 'y': [0]}

    b = CircleBoundaryConstraint([2, 2], 2)
    tf = TopFarmProblem(
        desvar,
        DummyCost(optimal, 'xy'),
        constraints=[b])
    _, state, _ = tf.optimize()

    npt.assert_array_less((state['x'] - 2)**2 + (state['y'] - 2)**2, 9)
    npt.assert_array_less(.9999999, state['x'])


def test_TopFarmProblem_with_cirleboundary_plot():
    if os.name == 'posix' and "DISPLAY" not in os.environ:
        pytest.xfail("No display")
    optimal = np.array([(0, 0)])
    desvar = dict(zip('xy', optimal.T))
    plot_comp = PlotComp()
    b = CircleBoundaryConstraint([1, 2], 3)
    tf = TopFarmProblem(
        desvar,
        DummyCost(optimal, 'xy'),
        constraints=[b],
        plot_comp=plot_comp)
    tf.evaluate()


def test_TopFarmProblem_with_cirleboundary_gradients():
    optimal = np.array([(0, 0)])
    desvar = dict(zip('xy', optimal.T + 1.5))
    plot_comp = NoPlot()
    b = CircleBoundaryConstraint([1, 2], 3)
    tf = TopFarmProblem(
        desvar,
        DummyCost(optimal, 'xy'),
        constraints=[b],
        plot_comp=plot_comp,
        driver=SimpleGADriver())
    tf.check_gradients(True)


def test_check_gradients():
    center = [0.1, 0]
    points = np.array([(-2, 0), (-1, 0), (-.5, 0), (.5, 0), (1, 0), (2, 0),
                       (0, -2), (0, -1), (0, -.5), (0, .5), (0, 1), (0, 2),
                       (-1, -.5), (-.5, -.25), (.5, .25), (1, .5),
                       (-1, .5), (-.5, .25), (.5, -.25), (1, -.5)
                       ])
    cb = CircleBoundaryComp(3, center, 1)
    d = cb.distances(*points.T)
    np.testing.assert_array_almost_equal(d, 1 - np.sqrt(((points - center)**2).sum(1)))
    dx, dy = [np.diagonal(d) for d in cb.gradients(*points.T)]
    eps = 1e-7
    d1 = cb.distances(points[:, 0] + eps, points[:, 1])
    np.testing.assert_array_almost_equal((d1 - d) / eps, dx)
    d2 = cb.distances(points[:, 0], points[:, 1] + eps)
    np.testing.assert_array_almost_equal((d2 - d) / eps, dy)


def test_move_inside():
    pbc = CircleBoundaryComp(1, (2, 1), 3)
    x0, y0 = [3, 3, 3, 12, 12, 12], [3, 5, 10, 8, 10, 12]
    state = pbc.satisfy({'x': x0, 'y': y0})
    x, y = state['x'], state['y']
    if 0:
        import matplotlib.pyplot as plt
        b = np.r_[pbc.xy_boundary, pbc.xy_boundary[:1]]
        plt.plot(b[:, 0], b[:, 1], 'k')
        for x0_, x_, y0_, y_ in zip(x0, x, y0, y):
            plt.plot([x0_, x_], [y0_, y_], '.-')
        plt.show()
    npt.assert_array_less((x - 2)**2 + (y - 1)**2, 3**2)



import matplotlib
from openmdao.api import ExplicitComponent
import matplotlib.pyplot as plt
import numpy as np
import topfarm
import sys


def mypause(interval):
    # pause without show
    backend = plt.rcParams['backend']
    if backend in matplotlib.rcsetup.interactive_bk:
        figManager = matplotlib._pylab_helpers.Gcf.get_active()
        if figManager is not None:
            canvas = figManager.canvas
            if canvas.figure.stale:
                canvas.draw()
            canvas.start_event_loop(interval)
            return


class NoPlot():
    """Plotting component for no plotting"""

    def show(self):
        pass


class XYPlotComp(ExplicitComponent):
    """Plotting component for turbine locations"""
    # colors = ['b', 'r', 'm', 'c', 'g', 'y', 'orange', 'indigo', 'grey'] * 100
    colors = [c['color'] for c in iter(matplotlib.rcParams['axes.prop_cycle'])] * 100

    def __init__(self, memory=10, delay=0.001, plot_initial=True, plot_improvements_only=False, ax=None, legendloc=1):
        """Initialize component for plotting turbine locations

        Parameters
        ----------
        memory : int, optional
            Number of previous iterations to remember
        delay : float, optional
            Time delay in seconds between plotting updates
        plot_initial : bool, optional
            Flag to plot the initial turbine locations
        plot_improvements_only : bool, optional
            Flag to plot only improvements in cost
        ax : matplotlib axes, optional
            Axes into which to make the plot
        legendloc : int
            Location of the legend in the plot
        """
        ExplicitComponent.__init__(self)
        self.delay = delay
        self.plot_improvements_only = plot_improvements_only
        self._ax = ax
        self.memory = memory
        self.delay = max([delay, 1e-6])
        self.plot_initial = plot_initial
        self.history = []
        self.counter = 0
        self.by_pass = False
        self.legendloc = legendloc

    @property
    def ax(self):
        return self._ax or plt.gca()

    def show(self):
        plt.show()

    def setup(self):
        if topfarm.x_key in self.problem.design_vars:
            units_x = self.problem.design_vars[topfarm.x_key][-1]
        else:
            units_x = None
        if topfarm.y_key in self.problem.design_vars:
            units_y = self.problem.design_vars[topfarm.y_key][-1]
        else:
            units_y = None
        self.add_input(topfarm.x_key, np.zeros(self.n_wt), units=units_x)
        self.add_input(topfarm.y_key, np.zeros(self.n_wt), units=units_y)
        if hasattr(self.problem, 'xy_boundary'):
            self.xy_boundary = self.problem.xy_boundary
        if hasattr(self.problem.cost_comp, 'output_key'):
            self.cost_key = self.problem.cost_comp.output_key
            self.cost_unit = self.problem.cost_comp.output_unit
        else:
            self.cost_key = "Cost"
            self.cost_unit = ""
        self.add_input(self.cost_key, 0.)
        self.add_output('plot_counter')

    def init_plot(self, limits):
        self.ax.cla()
        self.ax.axis('equal')

        mi = limits.min(0)
        ma = limits.max(0)
        ra = ma - mi + 1
        ext = .1
        xlim, ylim = np.array([mi - ext * ra, ma + ext * ra]).T
        self.ax.set_xlim(xlim)
        self.ax.set_ylim(ylim)

#     def plot_boundary(self):
#         b = np.r_[self.xy_boundary[:], self.xy_boundary[:1]]
#         plt.plot(b[:, 0], b[:, 1], 'k')

    def plot_constraints(self):
        for constr in self.problem.model.constraint_components:
            constr.plot(self.ax)

    def plot_history(self, x, y):
        rec = self.problem.recorder
        if rec.num_cases > 0:
            def get(xy, xy_key, pw):
                rec_xy = pw[xy_key][-self.memory:]
                if len(rec_xy.shape) == 1:
                    rec_xy = rec_xy[:, np.newaxis]
                return np.r_[rec_xy, [xy]]
            pw = self.problem.get_vars_from_recorder()
            x = get(x, topfarm.x_key, pw)
            y = get(y, topfarm.y_key, pw)
            for c, x_, y_ in zip(self.colors, x.T, y.T):
                self.ax.plot(x_, y_, '--', color=c)

    def plot_initial2current(self, x0, y0, x, y):
        rec = self.problem.recorder
        if rec.num_cases > 0:
            pw = self.problem.get_vars_from_recorder()
            x0 = np.atleast_1d(pw['x0'])
            y0 = np.atleast_1d(pw['y0'])
            for c, x0_, y0_, x_, y_ in zip(self.colors, x0, y0, x, y):
                self.ax.plot(x0_, y0_, '>', markerfacecolor=c, markeredgecolor='k')
                self.ax.plot((x0_, x_), (y0_, y_), '-', color=c)
            self.ax.plot([], [], '>k', markerfacecolor="#00000000", markeredgecolor='k', label='Initial position')

    def plot_current_position(self, x, y):
        for c, x_, y_ in zip(self.colors, x, y):
            self.ax.plot(x_, y_, 'o', color=c, ms=5)
            self.ax.plot(x_, y_, 'xk', ms=4)
        self.ax.plot([], [], 'xk', label='Current position')

    def set_title(self, cost0, cost):
        rec = self.problem.recorder
        if hasattr(self.problem.cost_comp, 'inc_or_exp'):
            inc_or_exp = self.problem.cost_comp.inc_or_exp
        else:
            inc_or_exp = 1.0
        title = "%d) %s %f %s" % (rec.num_cases, self.cost_key, cost * inc_or_exp, self.cost_unit)
        if cost0 != 0:
            title += " (%+.2f%%)" % ((cost - cost0) / cost0 * 100)
        self.ax.set_title(title)

    def get_initial(self):
        rec = self.problem.recorder
        if rec.num_cases > 0:
            pw = self.problem.get_vars_from_recorder()
            cost0 = self.problem.recorder[self.cost_key][0]
            # cost0 = pw['cost0']
            return pw['x0'], pw['y0'], cost0

    def compute(self, inputs, outputs):
        if self.by_pass is False:
            cost = inputs[self.cost_key][0]

            if (self.plot_improvements_only and
                'cost' in self.problem.recorder.driver_iteration_dict and
                len(self.problem.recorder['cost']) and
                    cost > self.problem.recorder['cost'].min()):
                return

            # find limits

            if (topfarm.x_key in self.problem.design_vars and
                    isinstance(self.problem.design_vars[topfarm.x_key], tuple) and
                    len(self.problem.design_vars[topfarm.x_key]) == 4):
                min_x, max_x = self.problem.design_vars[topfarm.x_key][1:-1]
            else:
                min_x, max_x = min(inputs[topfarm.x_key]), max(inputs[topfarm.x_key])
            if (topfarm.y_key in self.problem.design_vars and
                    isinstance(self.problem.design_vars[topfarm.y_key], tuple) and
                    len(self.problem.design_vars[topfarm.y_key]) == 4):
                min_y, max_y = self.problem.design_vars[topfarm.y_key][1:-1]
            else:
                min_y, max_y = min(inputs[topfarm.y_key]), max(inputs[topfarm.y_key])

            self.init_plot(np.array([[min_x, min_y], [max_x, max_y]]))

            self.plot_constraints()

            initial = self.get_initial()

            x = inputs[topfarm.x_key]
            y = inputs[topfarm.y_key]
            if initial is not None:
                x0, y0, cost0 = initial
                if self.plot_initial:
                    self.plot_initial2current(x0, y0, x, y)
                if self.memory > 0:
                    self.plot_history(x, y)
            else:
                cost0 = cost

            self.plot_current_position(x, y)
            self.set_title(cost0, cost)
            self.ax.legend(loc=self.legendloc)

            if self.counter == 0:
                plt.pause(1e-6)
            mypause(self.delay)

            self.counter += 1
            outputs['plot_counter'] = self.counter


class PlotComp(XYPlotComp):
    def __init__(self, memory=10, delay=0.001, plot_initial=True, plot_improvements_only=False, ax=None):
        XYPlotComp.__init__(self, memory=memory, delay=delay, plot_initial=plot_initial,
                            plot_improvements_only=plot_improvements_only, ax=ax)
        sys.stderr.write("%s is deprecated. Use XYPlotComp instead\n" % self.__class__.__name__)


class TurbineTypePlotComponent(XYPlotComp):
    """Plotting component for turbine types"""
    colors = ['b', 'r', 'm', 'c', 'g', 'y', 'orange', 'indigo', 'grey'] * 100
    markers = np.array(list("123v^<>.o48spP*hH+xXDd|_"))

    def __init__(self, turbine_type_names, **kwargs):
        """Initialize component for plotting turbine types

        Parameters
        ----------
        turbine_type_names : list of str
            Names of turbine types for legend
        **kwargs : keyword arguments, optional
            Keyword arguments that can be passed into XYPlotComp
        """
        self.turbine_type_names = turbine_type_names
        XYPlotComp.__init__(self, **kwargs)

    def setup(self):
        XYPlotComp.setup(self)
        self.add_input(topfarm.type_key, np.zeros(self.n_wt, dtype=np.int))

    def compute(self, inputs, outputs):
        self.types = np.asarray(inputs[topfarm.type_key], dtype=np.int)
        XYPlotComp.compute(self, inputs, outputs)

    def init_plot(self, limits):
        XYPlotComp.init_plot(self, limits)
        for m, n in zip(self.markers, self.turbine_type_names):
            self.ax.plot([], [], m + 'k', label=n)
        self.ax.legend()

    def plot_current_position(self, x, y):
        for m, c, x_, y_ in zip(self.markers[self.types], self.colors, x, y):
            # self.ax.plot(x_, y_, 'o', color=c, ms=5)
            self.ax.plot(x_, y_, m + 'k', markeredgecolor=c, markeredgewidth=1, ms=8)



from urllib.request import urlopen, Request
from urllib.parse import urlencode
import time
from _collections import defaultdict
import matplotlib.pyplot as plt
import numpy as np


def report_result(name, method=''):
    def wrap1(f):
        def wrap(*args, **kwargs):
            t = time.time()
            res = f(*args, **kwargs)
            ref = 365.8015799324624
            cost = -list(res)[0]
            duration = time.time() - t
            print("%s improved AEP to %f.4fGWh (%+.4f%%) in %.3fs" % (name, cost, (cost - ref) / ref * 100, duration))
            url = "http://tools.windenergy.dtu.dk/topfarm_ex/insert_result.asp"
            values = {'user': name,
                      'time': duration,
                      'aep': cost,
                      'comment': method,
                      }
            data = urlencode(values).encode("utf-8")
            req = Request(url, data)
            urlopen(req)
            return res
        return wrap
    return wrap1


def plot_result():
    url = "http://tools.windenergy.dtu.dk/topfarm_ex/view_result.asp"
    req = Request(url)
    s = urlopen(req).read().decode()
    res_dict = defaultdict(list)
    for r in s.split("<br>")[:-1]:
        name, duration, aep, comment = r.split(";")
        res_dict[name].append((duration, aep, comment))
    fig = plt.figure(figsize=(25, 10))
    ref_aep = 365.8015799324624
    duration_lst = []
    for i, name in enumerate(sorted(res_dict.keys())):
        duration = np.array([res[0] for res in res_dict[name]], dtype=np.float)
        duration_lst.extend(duration)
        aep = np.array([res[1] for res in res_dict[name]], dtype=np.float)
        marker = "v^<>.ospP*X"[i // 10]
        plt.plot(duration, (aep - ref_aep) / ref_aep * 100, '^', label=name, marker=marker)
    plt.plot([min(duration_lst), max(duration_lst)], [0, 0], 'gray')
    best = (418.9244064 - ref_aep) / ref_aep * 100
    plt.plot([min(duration_lst), max(duration_lst)], [best, best], 'gray')
    plt.xlabel('Time [s]')
    plt.ylabel('AEP improvement [%]')
    plt.legend()
    return fig


def reset_results():
    url = "http://tools.windenergy.dtu.dk/topfarm_ex/reset_results.asp"
    req = Request(url)
    urlopen(req)


if __name__ == '__main__':
    fig = plot_result()
    plt.show()



# %% Import section.

"""
This module contains classes and functions needed to form the loads constraint.
"""

import warnings
import numpy as np

from sklearn.neural_network import MLPRegressor
from sklearn.preprocessing import StandardScaler, MinMaxScaler

import topfarm


# %% Generic stuff.

def compute_error(res_1, res_2):
    """
    Compute the error between two set of data.

    Parameters
    ----------
    res_1 : numpy array or pandas DataFrame
        The reference results.
    res_2 : numpy array or pandas DataFrame
        The new results.

    Returns
    -------
    abs_err : numpy array or pandas DataFrame
        The absolute error between the results.
    rel_err : numpy array or pandas DataFrame
        The relative error between the results.

    """
    abs_err = res_2 - res_1
    rel_err = res_2 / res_1 - 1.0
    return abs_err, rel_err


# Dictionary used by predict_output() to call the correct function for
# each model type.
_switch_output = {}

# Dictionary used by predict_gradient() to call the correct function for
# each model type.
_switch_gradient = {}


# %% Add python functions to the switch dictionaries.

def _predict_output_python_function(model, input):
    """
    Predict output function for python functions.
    """
    output = model(input)
    return output


def _predict_gradient_python_function(model, input):
    """
    Predict gradient function for python functions.
    """
    output = model(input)
    return output


_switch_output[type(lambda: None)] = _predict_output_python_function
_switch_gradient[type(lambda: None)] = _predict_gradient_python_function


# %% Add scikit-learn neural networks to the switch dictionaries.

def _predict_output_scikit_MLPRegressor(model, input):
    """
    Predict output function for scikit-learn MLPRegressor objects.
    """
    output = model.predict(input)
    # Ensure that output is 2D even when there is only one output channel.
    if output.ndim == 1:
        output = output.reshape(-1, 1)
    return output


_switch_output[MLPRegressor] = _predict_output_scikit_MLPRegressor


# %% Add plugins to the switch dictionaries.

if 'openturnsloads' in topfarm.plugins:

    from openturnsloads.load import update_switch
    _switch_output_update, _switch_gradient_update = update_switch()
    _switch_output.update(_switch_output_update)
    _switch_gradient.update(_switch_gradient_update)


if 'tensorflowloads' in topfarm.plugins:

    from tensorflowloads.load import update_switch
    _switch_output_update, _switch_gradient_update = update_switch()
    _switch_output.update(_switch_output_update)
    _switch_gradient.update(_switch_gradient_update)

if 'wind2loads' in topfarm.plugins:

    from wind2loads.load import update_switch
    _switch_output_update, _switch_gradient_update = update_switch()
    _switch_output.update(_switch_output_update)
    _switch_gradient.update(_switch_gradient_update)


# %% Functions for evaluating surrogate models.

unknown_model_exception = KeyError(
    'Unknown model type. Please extend _switch_output '
    'with the appropriate predict function.')


def predict_output(model,
                   input,
                   model_in_keys=None,
                   input_scaler=None,
                   output_scaler=None,
                   boundary=None):
    """
    Predict the response of a model.

    Parameters
    ----------
    model : python function
            scikit-learn MLPRegressor
        The model to be evaluated, which can be a Multiple Input Multiple Output.
        Support for additional model types is provided through the Loads cutting
        edge plugins. model must return a 2D array, where each row is a
        different sample, and each column a different output.

    input : numpy.ndarray
            dict
        Input on which evaluate the model. If it is a 2D array, then it is
        assumed that each row is a different sample, and each column a
        different input variable. If instead it is a dictionary, then its
        keys are compared against the ones in model_in_keys, and a 2D
        array is assembled afterwards.

    model_in_keys : list
        List of input variables of the surrogate model. This variable is
        used only if input is a dictionary.

    input_scaler : sklearn.preprocessing.StandardScaler
                   sklearn.preprocessing.MinMaxScaler
        The model might have been constructed on a scaled input. This
        parameter ensures that input is scaled before evaluating the model.

    output_scaler : sklearn.preprocessing.StandardScaler
                    sklearn.preprocessing.MinMaxScaler
        The model might have been constructed on a scaled output. This
        parameter ensures that the predicted output is scaled back to its
        original dimensions.

    boundary : function
        A function that takes as input a 2D numpy array and retuns a list,
        with True if a point lies within the boudary, and False otherwise. A
        warning is issued if a point lies outside of the boundary.

    Returns
    -------
    output : numpy.ndarray
        Model output, optionally scaled through output_scaler.
        2D array, where each row is a different sample, and each column a
        different output.

    extrapolation_sample : list
        Identifiers of the points outside of the boundary.


    Raises
    ------
    KeyError : if the model is not supported.
    TypeError : if input is not a dictionary or a numpy array.
    Warning: if some points are outside of the boundary.

    """

    # Form the input array.
    if type(input) is dict:
        input_array = np.column_stack(
            [input[key] for key in model_in_keys])
    elif type(input) is np.ndarray:
        input_array = input
    else:
        raise TypeError(
            'Parameter input must be a dictionary or a numpy array.')

    # Check for extrapolation.
    if boundary is not None:
        extrapolation_sample = np.where(~boundary(input_array))[0].tolist()
    else:
        extrapolation_sample = []
    if extrapolation_sample:
        warnings.warn('Surrogate evaluated outside of input domain. See extrapolation_sample.')

    # If possible scale the input.
    if input_scaler is not None:
        input_scaler.copy = True
        input_scaled = input_scaler.transform(input_array)
    else:
        input_scaled = input_array

    # Predict the output by means of a switch on the model type.
    try:
        output = np.array(
            _switch_output[type(model)](model, input_scaled))
    except KeyError:
        # Unknown model.
        raise unknown_model_exception

    # If possible scale back the output.
    if output_scaler is not None:
        output = output_scaler.inverse_transform(output)

    return output, extrapolation_sample


def predict_gradient(model,
                     input,
                     model_in_keys=None,
                     input_scaler=None,
                     output_scaler=None,
                     boundary=None):
    """
    Predict the gradient of a model.

    Parameters
    ----------
    model : python function
            scikit-learn MLPRegressor
        The model to be evaluated, which must be a Multiple Input Single Output.
        Support for additional model types is provided through the Loads cutting
        edge plugins.

    input : numpy.ndarray
            dict
        Input on which evaluate the model. If it is a 2D array, then it is
        assumed that each row is a different sample, and each column a
        different input variable. If instead it is a dictionary, then its
        keys are compared against the ones in model_in_keys, and a 2D
        array is assembled afterwards.

    model_in_keys : list
        List of input variables of the surrogate model. This variable is
        used only if input is a dictionary.

    input_scaler : sklearn.preprocessing.StandardScaler
                   sklearn.preprocessing.MinMaxScaler
        The model might have been constructed on a scaled input.
        The derivative is computed as
            y' = f'(x)
               = ft'(xt) * g'(x)
        with f the unscaled model, ft() the scaled one, x the unscaled input,
        xt the scaled input and g() the input_scaler.

    output_scaler : sklearn.preprocessing.StandardScaler
                    sklearn.preprocessing.MinMaxScaler
        The model might have been constructed on a scaled output. This
        parameter ensures that the predicted gradient is scaled back to its
        original dimensions. The derivative is computed as
            y' = f'(x)
               = h^{-1}'(yt) * ft'(x)
        with h() the output_scaler, yt the scaled output, ft() the scaled
        model and x the unscaled input. If there are both input and output
        scalares, the formula becomes
            y' = h^{-1}'(yt) * ft'(xt) * g'(x)

    Returns
    -------
    gradient : numpy.ndarray
        Model gradient, optionally scaled through output_scaler.
        2D array, where each row is associated to a different sample.

    extrapolation_sample : list
        Identifiers of the points outside of the boundary.

    Raises
    ------
    KeyError : if the model is not supported.
    TypeError : if input is not a dictionary or a numpy array.
    Warning: if some points are outside of the boundary.

    """

    # This function can be extended to the Multiple Output case,
    # but we must be careful with the shape of the output:
    # - always 2D?
    # - 2D or 3D depending on the case?

    # Form the input array.
    if type(input) is dict:
        input_array = np.column_stack(
            [input[key] for key in model_in_keys])
    elif type(input) is np.ndarray:
        input_array = input
    else:
        raise TypeError(
            'Parameter input must be a dictionary or a numpy array.')

    # Check for extrapolation.
    if boundary is not None:
        extrapolation_sample = np.where(~boundary(input_array))[0].tolist()
    else:
        extrapolation_sample = []
    if extrapolation_sample:
        warnings.warn('Surrogate evaluated outside of input domain. See extrapolation_sample.')

    # If possible scale the input.
    if input_scaler is not None:
        input_scaler.copy = True
        input_scaled = input_scaler.transform(input_array)
        try:
            gradient = np.array(
                _switch_gradient[type(model)](model, input_scaled))
            if type(input_scaler) is MinMaxScaler:
                gradient *= input_scaler.scale_
            elif ((type(input_scaler) is StandardScaler) and
                  (input_scaler.scale_ is not None)):
                gradient /= input_scaler.scale_
        except KeyError:
            raise unknown_model_exception
    else:  # The input is not scaled.
        try:
            gradient = np.array(
                _switch_gradient[type(model)](model, input_array))
        except KeyError:
            raise unknown_model_exception

    # If possible scale back the gradient.
    if output_scaler is not None:
        if type(output_scaler) is MinMaxScaler:
            gradient /= output_scaler.scale_
        elif ((type(output_scaler) is StandardScaler) and
              (output_scaler.scale_ is not None)):
            gradient *= output_scaler.scale_

    return gradient, extrapolation_sample


# %% Classes needed for the load constraint.

class SurrogateModel():
    """
    Class to store variables related to a surrogate model.
    """

    def __init__(self,
                 model=None,
                 input_scaler=None,
                 output_scaler=None,
                 input_channel_names=None,
                 output_channel_name=None,
                 ):
        """
        Initialize SurrogateModel.

        Parameters
        ----------
        model : any surrogate model supported by predict_output().
            The surrogate model.

        input_scaler : sklearn.preprocessing.StandardScaler
                       sklearn.preprocessing.MinMaxScaler
            The scaler used to transform the input.

        output_scaler : sklearn.preprocessing.StandardScaler
                       sklearn.preprocessing.MinMaxScaler
            The scaler used to anti-transform the output.

        input_channel_names : list of strings
            The name of the input channels.

        output_channel_name : str
            The name of the input channel.
        """
        self.model = model
        self.input_scaler = input_scaler
        self.output_scaler = output_scaler
        self.input_channel_names = input_channel_names
        self.output_channel_name = output_channel_name




"""TOPFARM OpenMDAO Problem.

This module contains the OpenMDAO problem that can be used for
different types of optimization problems.

Notes
-----

A new TopFarmProblem must be constructed for every optimization problem that
has different design variables. For example, if one problem optimizes a
wind-farm layout but another optimizes turbine types at fixed turbine
positions, these require OpenMDAO problems that are constructed differently.

To get around this, TOPFARM has a base optimization problem,
``TopFarmProblem`` that inherits from the ``Problem`` class in OpenMDAO.

To disable MPI for topfarm add this before importing topfarm
```from openmdao.utils import mpi
mpi.MPI = None```
"""
import time
import numpy as np
import warnings
import copy
from openmdao.api import Problem, IndepVarComp, Group, ParallelGroup,\
    ExplicitComponent, ListGenerator, DOEDriver, SimpleGADriver
from openmdao.drivers.doe_generators import DOEGenerator
from openmdao.utils import mpi
import topfarm
from topfarm.recorders import NestedTopFarmListRecorder,\
    TopFarmListRecorder, split_record_id
from topfarm.plotting import NoPlot
from topfarm.easy_drivers import EasyScipyOptimizeDriver, EasySimpleGADriver, EasyDriverBase
from topfarm.utils import smart_start
from topfarm.constraint_components.spacing import SpacingComp
from topfarm.constraint_components.boundary import BoundaryBaseComp
from topfarm.constraint_components.penalty_component import PenaltyComponent, PostPenaltyComponent
from topfarm.cost_models.aggregated_cost import AggregatedCost


class TopFarmBaseGroup(Group):
    def __init__(self, comps=[], output_key=None, output_unit=''):
        super().__init__()
        self.comps = []
        self.obj_comp = None
        for i, comp in enumerate(comps):
            if hasattr(comp, 'objective') and comp.objective:
                self.output_key = comp.output_key
                self.output_unit = comp.output_unit
                self.cost_factor = comp.cost_factor
                self.obj_comp = comp
            else:
                self.comps.append(comp)


class TopFarmGroup(TopFarmBaseGroup):
    def __init__(self, comps=[], output_key=None, output_unit=''):
        super().__init__(comps, output_key, output_unit)
        for i, comp in enumerate(comps):
            self.add_subsystem('comp_{}'.format(i), comp, promotes=['*'])


class TopFarmParallelGroup(TopFarmBaseGroup):
    def __init__(self, comps=[], output_key=None, output_unit=''):
        super().__init__(comps, output_key, output_unit)
        parallel = ParallelGroup()
        for i, comp in enumerate(self.comps):
            parallel.add_subsystem('comp_{}'.format(i), comp, promotes=['*'])
        self.add_subsystem('parallel', parallel, promotes=['*'])
        self.add_subsystem('objective', self.obj_comp, promotes=['*'])


class TopFarmProblem(Problem):

    def __init__(self, design_vars, cost_comp=None, driver=EasyScipyOptimizeDriver(),
                 constraints=[], plot_comp=NoPlot(), record_id=None,
                 expected_cost=1, ext_vars={}, post_constraints=[], approx_totals=False, additional_recorders=None):
        """Initialize TopFarmProblem

        Parameters
        ----------
        design_vars : dict or list of key-initial_value-tuples
            Design variables for the problem.\n
            Ex: {'x': [1,2,3], 'y':([3,2,1],0,1), 'z':([4,5,6],[4,5,4], [6,7,6])}\n
            Ex: [('x', [1,2,3]), ('y',([3,2,1],0,1)), ('z',([4,5,6],[4,5,4], [6,7,6]))]\n
            Ex: [('x', ([1,2,3],0,3,'m')), ('y',([3,2,1],'m')), ('z',([4,5,6],[4,5,4], [6,7,6]))]\n
            Ex: zip('xy', pos.T)\n
            The keys (x, y, z) are the names of the design variable.\n
            The values are either\n
            - the initial value or\n
            - on of the following tuples:
                (initial value, unit)
                (initial value, lower bound, upper bound)
                (initial value, lower bound, upper bound, unit)
        cost_comp : ExplicitComponent or TopFarmProblem or TopFarmGroup
            Component that provides the cost function. It has to be the style
            of an OpenMDAO v2 ExplicitComponent.
            Pure python cost functions can be wrapped using ``CostModelComponent``
            class in ``topfarm.cost_models.cost_model_wrappers``.\n
            ExplicitComponent are wrapped into a TopFarmGroup.\n
            For nested problems, the cost comp_comp is typically a TopFarmProblem
        driver : openmdao Driver, optinal
            Driver used to solve the optimization driver. For an example, see the
            ``EasyScipyOptimizeDriver`` class in ``topfarm.easy_drivers``.
        constraints : list of Constraint-objects
            E.g. XYBoundaryConstraint, SpacingConstraint
        plot_comp : ExplicitComponent, optional
            OpenMDAO ExplicitComponent used to plot the state (during
            optimization).
            For no plotting, pass in the ``topfarm.plotting.NoPlot`` class.
        record_id : string "<record_id>:<case>", optional
            Identifier for the optimization. Allows a user to restart an
            optimization where it left off.\n
            record_id can be name (saves as recordings/<name>.pkl), abs or relative path
            Case can be:\n
            - "", "latest", "-1": Continue from latest\n
            - "best": Continue from best case (minimum cost)\n
            - "0": Start from scratch (initial position)\n
            - "4": Start from case number 4\n
        expected_cost : int, float or None, optional
            Used to scale the cost, default is 1. This has influence on some drivers, e.g.
            SLSQP where it affects the step size\n
            If None, the value is found by evaluating the cost function
        ext_vars : dict or list of key-initial_value tuple
            Used for nested problems to propagate variables from parent problem\n
            Ex. {'type': [1,2,3]}\n
            Ex. [('type', [1,2,3])]\n
        post_constraints : list of Constraint-objects that needs the cost component to be
            evaluated, unlike (pre-)constraints which are evaluated before the cost component.
            E.g. LoadConstraint
        approx_totals : bool or dict
            If True, approximates the total derivative of the cost_comp group,
            skipping the partial ones. If it is a dictionary, it's elements
            are passed to the approx_totals function of an OpenMDAO Group.
        additional_recorders: list(Recorder) or None
            A list of additional recorders to be added to the problem

        Examples
        --------
        See main() in the bottom of this file
        """
        if mpi.MPI:
            comm = None
        else:
            from openmdao.utils.mpi import FakeComm
            comm = FakeComm()

        self._additional_recorders = additional_recorders

        Problem.__init__(self, comm=comm)
        if cost_comp:
            if isinstance(cost_comp, TopFarmProblem):
                cost_comp = cost_comp.as_component()
            elif isinstance(cost_comp, ExplicitComponent) and (len(post_constraints) > 0):
                cost_comp = TopFarmGroup([cost_comp])
            cost_comp.parent = self
        self.cost_comp = cost_comp

        if isinstance(driver, list):
            driver = DOEDriver(ListGenerator(driver))
        elif isinstance(driver, DOEGenerator):
            driver = DOEDriver(generator=driver)
        self.driver = driver
        self.driver.recording_options['record_desvars'] = True
        self.driver.recording_options['includes'] = ['*']
        self.driver.recording_options['record_inputs'] = True

        self.plot_comp = plot_comp

        self.record_id = record_id
        self.load_recorder()
        if not isinstance(approx_totals, dict) and approx_totals:
            approx_totals = {'method': 'fd'}

        if not isinstance(design_vars, dict):
            design_vars = dict(design_vars)
        self.design_vars = design_vars
        self.indeps = self.model.add_subsystem('indeps', IndepVarComp(), promotes=['*'])
        for k, v in design_vars.items():
            if isinstance(v, tuple):
                if (not isinstance(v[-1], str)) or (not v[-1]):
                    design_vars[k] += (None, )
            else:
                design_vars[k] = (design_vars[k], None)
            v = design_vars[k]
            self.indeps.add_output(k, v[0], units=v[-1])

        for k in [topfarm.x_key, topfarm.y_key, topfarm.type_key]:
            if k in design_vars:
                self.n_wt = len(design_vars[k][0])
                break
        else:
            self.n_wt = 0

        constraints_as_penalty = ((not self.driver.supports['inequality_constraints'] or
                                  isinstance(self.driver, SimpleGADriver) or
                                  isinstance(self.driver, EasySimpleGADriver)) and
                                  len(constraints) + len(post_constraints) > 0)

        if len(constraints) > 0:
            self.model.add_subsystem('pre_constraints', ParallelGroup(), promotes=['*'])
            for constr in constraints:
                if constraints_as_penalty:
                    constr.setup_as_penalty(self)
                else:
                    constr.setup_as_constraint(self)
                    # Use the assembled Jacobian.
                    self.model.pre_constraints.options['assembled_jac_type'] = 'csc'
                    self.model.pre_constraints.linear_solver.assemble_jac = True
            penalty_comp = PenaltyComponent(constraints, constraints_as_penalty)
            self.model.add_subsystem('penalty_comp', penalty_comp, promotes=['*'])

        self.model.constraint_components = [constr.constraintComponent for constr in constraints]

        for k, v in design_vars.items():
            if isinstance(driver, EasyDriverBase):
                kwargs = driver.get_desvar_kwargs(self.model, k, v)
            else:
                kwargs = EasyDriverBase.get_desvar_kwargs(None, self.model, k, v)
            self.model.add_design_var(k, **kwargs)

        for k, v in ext_vars.items():
            self.indeps.add_output(k, v)
        self.ext_vars = ext_vars

        if cost_comp:
            self.model.add_subsystem('cost_comp', cost_comp, promotes=['*'])

            if expected_cost is None:
                expected_cost = self.evaluate()[0]
                self._setup_status = 0
            if isinstance(driver, EasyDriverBase) and driver.supports_expected_cost is False:
                expected_cost = 1
            if isinstance(cost_comp, Group) and approx_totals:
                cost_comp.approx_totals(**approx_totals)
            # Use the assembled Jacobian.
            if 'assembled_jac_type' in self.model.cost_comp.options:
                self.model.cost_comp.options['assembled_jac_type'] = 'dense'
                self.model.cost_comp.linear_solver.assemble_jac = True

        else:
            self.indeps.add_output('cost')

        if len(post_constraints) > 0:
            if constraints_as_penalty:
                penalty_comp = PostPenaltyComponent(post_constraints, constraints_as_penalty)
                self.model.add_subsystem('post_penalty_comp', penalty_comp, promotes=['*'])
            else:
                for constr in post_constraints:
                    self.model.add_constraint(constr[0],
                                              upper=np.full(self.n_wt, constr[1]))
                    # Use the assembled Jacobian.
#                    self.model.cost_comp.post_constraints.options['assembled_jac_type'] = 'csc'
#                    self.model.cost_comp.post_constraints.linear_solver.assemble_jac = True

        aggr_comp = AggregatedCost(constraints_as_penalty, constraints, post_constraints)
        self.model.add_subsystem('aggr_comp', aggr_comp, promotes=['*'])
        self.model.add_objective('aggr_cost', scaler=1 / abs(expected_cost))

        if plot_comp and not isinstance(plot_comp, NoPlot):
            self.model.add_subsystem('plot_comp', plot_comp, promotes=['*'])
            plot_comp.problem = self
            plot_comp.n_wt = self.n_wt

        self.setup()

    @property
    def cost(self):
        return self['aggr_cost'][0]

    def __getitem__(self, name):
        return Problem.__getitem__(self, name).copy()

    @property
    def state(self):
        """Return current state"""
        self.setup()
        state = {k: self[k] for k in self.design_vars}
        state.update({k: self[k] for k in self.ext_vars})
        if hasattr(self.cost_comp, 'state'):
            state.update(self.cost_comp.state)
        if hasattr(self.cost_comp, 'additional_output'):
            state.update({k: self[k] for k, _ in self.cost_comp.additional_output})
        return state

    def state_array(self, keys):
        self.setup()
        return np.array([self[k] for k in keys]).T

    def update_state(self, state):
        for k, v in state.items():
            try:
                c = self[k]  # fail if k not exists
                v = np.array(v)
                if hasattr(c, 'shape') and c.shape != v.shape:
                    v = v.reshape(c.shape)
                self[k] = v
            except KeyError:
                pass

    def load_recorder(self):
        if hasattr(self.cost_comp, 'problem'):
            self.recorder = NestedTopFarmListRecorder(self.cost_comp, self.record_id)
        else:
            self.recorder = TopFarmListRecorder(self.record_id)

    def get_vars_from_recorder(self):
        rec = self.recorder
        x = np.array(rec[topfarm.x_key])
        y = np.array(rec[topfarm.y_key])
        c = np.array(rec[topfarm.cost_key])
        x0 = x[0]
        y0 = y[0]
        cost0 = c[0]
        return {'x0': x0, 'y0': y0, 'cost0': cost0,
                topfarm.x_key: x, topfarm.y_key: y, 'c': c}

    def setup(self):
        if self._setup_status == 0:
            Problem.setup(self, check=True)
        if self._setup_status < 2:
            with warnings.catch_warnings():
                warnings.filterwarnings('error')
                try:
                    if len(self.driver._rec_mgr._recorders) == 0:
                        tmp_recorder = TopFarmListRecorder()
                        self.driver.add_recorder(tmp_recorder)
                        Problem.final_setup(self)
                    else:
                        Problem.final_setup(self)
                except Warning as w:
                    if str(w).startswith('Inefficient choice of derivative mode'):
                        Problem.setup(self, check=True, mode='fwd')
                    else:
                        raise w
                finally:
                    try:
                        self.driver._rec_mgr._recorders.remove(tmp_recorder)
                    except Exception:
                        pass

    def evaluate(self, state={}, disp=False):
        """Evaluate the cost model

        Parameters
        ----------
        state : dict, optional
            Initial state\n
            Ex: {'x': [1,2,3], 'y':[3,2,1]}\n
            The current state is used for unspecified variables
        disp : bool, optional
            if True, the time used for the optimization is printed

        Returns
        -------
        Current cost : float
        Current state : dict
        """
        tmp_recorder = TopFarmListRecorder()
        self.driver.add_recorder(tmp_recorder)
        self.setup()
        self.update_state(state)
        t = time.time()
        self.run_model()
        if disp:
            print("Evaluated in\t%.3fs" % (time.time() - t))
        self.driver._rec_mgr._recorders.remove(tmp_recorder)
        return self.cost, copy.deepcopy(self.state)

    def evaluate_gradients(self, disp=False):
        """Evaluate the gradients."""
        self.setup()
        t = time.time()
        rec = TopFarmListRecorder()
        self.driver.add_recorder(rec)
        res = self.compute_totals(['aggr_cost'], wrt=[topfarm.x_key, topfarm.y_key], return_format='dict')
        self.driver._rec_mgr._recorders.remove(rec)
        if disp:
            print("Gradients evaluated in\t%.3fs" % (time.time() - t))
        return res

    def optimize(self, state={}, disp=False):
        """Run the optimization problem

        Parameters
        ----------
        state : dict, optional
            Initial state\n
            Ex: {'x': [1,2,3], 'y':[3,2,1]}\n
            The current state is used to unspecified variables
        disp : bool, optional
            if True, the time used for the optimization is printed

        Returns
        -------
        Optimized cost : float
        state : dict
        recorder : TopFarmListRecorder or NestedTopFarmListRecorder
        """
        self.load_recorder()
        self.update_state(state)
        if self.recorder.num_cases > 0:
            try:
                self.update_state({k: self.recorder[k][-1] for k in self.state.keys() if k not in state})
            except (ValueError, KeyError):
                # Restart optimize with n
                self.record_id = split_record_id(self.record_id)[0] + ":0"
                return self.optimize(state, disp)

        self.driver.add_recorder(self.recorder)
        if self._additional_recorders:
            for r in self._additional_recorders:
                self.driver.add_recorder(r)

#        self.recording_options['includes'] = ['*']
#        self.driver.recording_options['record_desvars'] = True
#        self.driver.recording_options['includes'] = ['*']
#        self.driver.recording_options['record_inputs'] = True
        self.setup()
        t = time.time()
        self.run_driver()
        self.cleanup()
        if disp:
            print("Optimized in\t%.3fs" % (time.time() - t))
        if self.driver._rec_mgr._recorders != []:  # in openmdao<2.4 cleanup does not delete recorders
            self.driver._rec_mgr._recorders.remove(self.recorder)
        if isinstance(self.driver, DOEDriver) or isinstance(self.driver, SimpleGADriver):
            costs = self.recorder['aggr_cost']
            best_case_index = int(np.argmin(costs))
            best_state = {k: self.recorder[k][best_case_index] for k in self.design_vars}
            self.evaluate(best_state)
        return self.cost, copy.deepcopy(self.state), self.recorder

    def check_gradients(self, check_all=False, tol=1e-3):
        """Check gradient computations"""
        self.setup()
        if check_all:
            comp_name_lst = [comp.pathname for comp in self.model.system_iter()
                             if hasattr(comp, '_has_compute_partials') and comp._has_compute_partials]
        else:
            comp_name_lst = [self.cost_comp.pathname]
        print("checking %s" % ", ".join(comp_name_lst))
        res = self.check_partials(includes=comp_name_lst, compact_print=True)
        for comp in comp_name_lst:
            var_pair = [(x, dx) for x, dx in res[comp].keys()
                        if (x not in ['cost_comp_eval'] and
                            not x.startswith('penalty') and
                            not dx.startswith('penalty'))]
            worst = var_pair[np.argmax([res[comp][k]['rel error'].forward for k in var_pair])]
            err = res[comp][worst]['rel error'].forward
            if err > tol:
                raise Warning("Mismatch between finite difference derivative of '%s' wrt. '%s' and derivative computed in '%s' is: %f" %
                              (worst[0], worst[1], comp, err))

    def as_component(self):
        return ProblemComponent(self)

    def get_DOE_list(self):
        self.setup()
        assert isinstance(
            self.driver, DOEDriver), 'get_DOE_list only applies to DOEDrivers, and the current driver is: %s' % type(self.driver)
        case_gen = self.driver.options['generator']
        return [c for c in case_gen(self.model.get_design_vars(recurse=True), self.model)]

    def get_DOE_array(self):
        return np.array([[v for _, v in c] for c in self.get_DOE_list()])

    @property
    def turbine_positions(self):
        return np.array([self[k] for k in [topfarm.x_key, topfarm.y_key]]).T

    def smart_start(self, XX, YY, ZZ, radius=None):
        assert XX.shape == YY.shape
        if len(XX.shape) == 1:
            XX, YY = np.meshgrid(XX, YY)
        ZZ_is_func = hasattr(ZZ, '__call__')
        spacing_comp_lst = [c for c in self.model.constraint_components if isinstance(c, SpacingComp)]
        if len(spacing_comp_lst) == 1:
            min_spacing = spacing_comp_lst[0].min_spacing
        else:
            min_spacing = 0
        X, Y = XX.flatten(), YY.flatten()
        if not ZZ_is_func:
            Z = ZZ.flatten()
        else:
            Z = ZZ
        for comp in self.model.constraint_components:
            if isinstance(comp, BoundaryBaseComp):
                dist = comp.distances(X, Y)
                if len(dist.shape) == 2:
                    dist = dist.min(1)
                mask = dist >= 0
                X, Y = X[mask], Y[mask]
                if not ZZ_is_func:
                    Z = Z[mask]
        x, y = smart_start(X, Y, Z, self.n_wt, min_spacing, radius)
        self.update_state({topfarm.x_key: x, topfarm.y_key: y})
        return x, y


class ProblemComponent(ExplicitComponent):
    """class used to wrap a TopFarmProblem as a cost_component"""

    def __init__(self, problem, additional_inputs=[]):
        ExplicitComponent.__init__(self)
        self.problem = problem
        self.additional_inputs = additional_inputs

    def setup(self):
        missing_in_problem = (set([c[0] for c in self.parent.indeps._indep_external]) -
                              set([c[0] for c in self.problem.indeps._indep_external]))
        self.missing_attrs = []
        for name, val, kwargs in self.parent.indeps._indep_external:
            self.add_input(name, val=val, **{k: kwargs[k] for k in ['units']})
            self.missing_attrs.append(name)
            if name in missing_in_problem:
                self.problem.indeps.add_output(name, val, **kwargs)
        self.problem._setup_status = 0  # redo initial setup
        self.problem.setup()

        self.add_output('cost', val=0.0)
        if hasattr(self.problem.cost_comp, "output_key"):
            self.add_output(self.problem.cost_comp.output_key, val=0.0)
        self.comp_lst = [comp for comp in self.problem.model.system_iter()]

    @property
    def state(self):
        return self.problem.state

    def cost_function(self, **kwargs):
        return self.problem.optimize(kwargs)[0]

    def set_input_as_option(self, inputs):
        for comp in self.comp_lst:
            for attr in self.missing_attrs:
                if attr in comp.options:
                    comp.options[attr] = inputs[attr]

    def compute(self, inputs, outputs):
        self.set_input_as_option(inputs)
        outputs['cost'] = self.cost_function(**inputs)
        if hasattr(self.problem.cost_comp, "output_key"):
            output_key = self.problem.cost_comp.output_key
            outputs[output_key] = self.problem[output_key]


def main():
    if __name__ == '__main__':
        from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp
        from topfarm.constraint_components.spacing import SpacingConstraint
        from topfarm.constraint_components.boundary import XYBoundaryConstraint
        from openmdao.api import view_model

        initial = np.array([[6, 0], [6, -8], [1, 1]])  # initial turbine layouts
        optimal = np.array([[2.5, -3], [6, -7], [4.5, -3]])  # optimal turbine layouts
        boundary = np.array([(0, 0), (6, 0), (6, -10), (0, -10)])  # turbine boundaries
        desired = np.array([[3, -3], [7, -7], [4, -3]])  # desired turbine layouts
        drivers = [EasySimpleGADriver(max_gen=10, pop_size=100, bits={'x': [12] * 3, 'y':[12] * 3}, random_state=1),
                   EasyScipyOptimizeDriver()]
        plot_comp = DummyCostPlotComp(optimal)
        tf = TopFarmProblem(
            design_vars=dict(zip('xy', initial.T)),
            cost_comp=DummyCost(optimal_state=desired, inputs=['x', 'y']),
            constraints=[XYBoundaryConstraint(boundary),
                         SpacingConstraint(2)
                         ],
            driver=drivers[1],
            plot_comp=plot_comp
        )
        cost, _, recorder = tf.optimize()
        plot_comp.show()


main()



import numpy as np
from topfarm.cost_models.dummy import DummyCost, DummyCostPlotComp

from topfarm.plotting import NoPlot
from openmdao.drivers.genetic_algorithm_driver import SimpleGADriver
from topfarm.tests import npt
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm import TopFarmProblem


initial = np.array([[0, 0, 0], [6, 0, 0], [6, -10, 0]])  # initial turbine layouts
optimal = np.array([[2.5, -3], [6, -7], [4.5, -3]])  # desired turbine layouts
boundary = np.array([(0, 0), (6, 0), (6, -10), (0, -10)])  # turbine boundaries
desired = np.array([[3, -3, 1], [7, -7, 2], [4, -3, 3]])  # desired turbine layouts


def test_setup_as_constraint_xy():
    # plot_comp = DummyCostPlotComp(desired)
    plot_comp = NoPlot()

    tf = TopFarmProblem(
        {'x': initial[:, 0], 'y': initial[:, 1]},
        DummyCost(desired[:, :2]),
        constraints=[XYBoundaryConstraint(boundary)],
        driver=EasyScipyOptimizeDriver(disp=False),
        plot_comp=plot_comp)

    tf.optimize()
    tb_pos = tf.turbine_positions[:, :2]
    tf.plot_comp.show()
    tol = 1e-4
    assert tb_pos[1][0] < 6 + tol  # check within border


def test_setup_as_constraint_z():
    tf = TopFarmProblem(
        {'z': (initial[:, 2], 0, 2)},
        DummyCost(desired[:, :2], 'z'),
        driver=EasyScipyOptimizeDriver(disp=False),
    )

    tf.optimize()
    npt.assert_array_less(tf['z'], 2 + 1e-10)


def test_setup_as_constraint_xyz():
    desvar = dict(zip('xy', initial.T))
    desvar['z'] = (initial[:, 2], 0, 2)
    tf = TopFarmProblem(
        desvar,
        DummyCost(desired, 'xyz'),
        driver=EasyScipyOptimizeDriver(disp=False),
        constraints=[XYBoundaryConstraint(boundary)])
    tf.optimize()
    tb_pos = tf.turbine_positions
    tol = 1e-4
    assert tb_pos[1][0] < 6 + tol  # check within border
    npt.assert_array_less(tf['z'], 2 + tol)  # check within height limit


def test_setup_as_penalty_xy():
    driver = SimpleGADriver()
    tf = TopFarmProblem(
        dict(zip('xy', initial.T)),
        DummyCost(desired),
        constraints=[XYBoundaryConstraint(boundary)],
        driver=driver)

    # check normal result if boundary constraint is satisfied
    assert tf.evaluate()[0] == 121
    # check penalized result if boundary constraint is not satisfied
    assert tf.evaluate({'x': [2.5, 7, 4.5], 'y': [-3., -7., -3.], 'z': [0., 0., 0.]})[0] == 1e10 + 1


def test_setup_as_penalty_none():
    driver = SimpleGADriver()
    tf = TopFarmProblem(
        dict(zip('xy', initial.T)),
        DummyCost(desired),
        driver=driver)

    # check that it does not fail if xy and z is not specified
    assert tf.evaluate()[0] == 121
    assert tf.evaluate({'x': [2.5, 7, 4.5], 'y': [-3., -7., -3.], 'z': [0., 0., 0.]})[0] == .5



'''
Created on 19/04/2018

@author: Mads
'''
import numpy as np


class WindResource(object):
    def __init__(self, f, a, k, ti):
        wdir = np.linspace(0, 360, len(f), endpoint=False)
        indexes = np.argmin((np.tile(np.arange(360), (len(f), 1)).T - wdir + (360 / len(f) / 2)) % 360, 1)
        self.f = np.array(f)[indexes] / (360 / len(f))
        self.a = np.array(a)[indexes]
        self.k = np.array(k)[indexes]
        self.ti = np.array(ti)[indexes]

    def weibull_weight(self, WS, A, k):
        def cdf(ws, A=A, k=k):
            return 1 - np.exp(-(ws / A) ** k)
        dWS = np.diff(WS[:2], 1, 0) / 2
        return cdf(WS + dWS) - cdf(WS - dWS)

    def __call__(self, turbine_positions, wdir, wsp):
        # f(turbine_positions, wdir, wsp) -> WS[nWT,nWdir,nWsp], TI[nWT,nWdir,nWsp), Weight[nWdir,nWsp]
        WD, WS = np.meshgrid(wdir, wsp)
        weight = self.weibull_weight(WS, self.a[WD], self.k[WD]) * self.f[wdir]
        WD, WS = np.tile(WD, (len(turbine_positions), 1, 1)), np.tile(WS, (len(turbine_positions), 1, 1))
        return WD, WS, self.ti[WD], weight


def main():
    if __name__ == '__main__':
        f = [0.035972, 0.039487, 0.051674, 0.070002, 0.083645, 0.064348, 0.086432, 0.117705, 0.151576, 0.147379, 0.10012, 0.05166]
        A = [9.176929, 9.782334, 9.531809, 9.909545, 10.04269, 9.593921, 9.584007, 10.51499, 11.39895, 11.68746, 11.63732, 10.08803]
        k = [2.392578, 2.447266, 2.412109, 2.591797, 2.755859, 2.595703, 2.583984, 2.548828, 2.470703, 2.607422, 2.626953, 2.326172]
        ti = np.zeros_like(f) + .1
        wr = WindResource(f, A, k, ti)

        print(wr(turbine_positions=(0, 0), wdir=[0, 30, 60], wsp=[4, 5]))


main()



from topfarm.easy_drivers import EasyRandomSearchDriver
from topfarm.tests import npt
from topfarm.tests.test_files import xy3tb
from topfarm.drivers.random_search_driver import RandomizeTurbinePosition
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint
import numpy as np


def test_TopFarmProblem():
    tf = xy3tb.get_tf(design_vars={'x': [3, 7, 4], 'y': [-3, -7, -3]},
                      constraints=[])

    cost, state, _ = tf.optimize()
    npt.assert_almost_equal(cost, 0)
    npt.assert_array_almost_equal(state['x'], xy3tb.desired[:, 0])
    npt.assert_array_almost_equal(state['y'], xy3tb.desired[:, 1])


def test_TopFarmProblemLimits():

    tf = xy3tb.get_tf(design_vars={'x': (xy3tb.initial[:, 0], -3, 3),
                                   'y': (xy3tb.initial[:, 1], [-4, -3, -2], [2, 3, 4])},
                      driver=EasyRandomSearchDriver(RandomizeTurbinePosition(1), max_iter=100),
                      constraints=[])
    tf.evaluate()
    desvars = tf.driver._designvars
    npt.assert_equal(desvars['indeps.x']['lower'], -3)
    npt.assert_equal(desvars['indeps.x']['upper'], 3)
    npt.assert_array_equal(desvars['indeps.y']['lower'], [-4, -3, -2])
    npt.assert_array_equal(desvars['indeps.y']['upper'], [2, 3, 4])


def test_TopFarmProblemSpacingConstraint():
    tf = xy3tb.get_tf(design_vars={'x': [3, 7, 4], 'y': [-3, -7, -3]},
                      constraints=[SpacingConstraint(2)])
    tf.evaluate({'x': xy3tb.desired[:, 0], 'y': xy3tb.desired[:, 1]})
    npt.assert_array_equal(tf['wtSeparationSquared'], [32, 1, 25])

    _, state, _ = tf.optimize()
    npt.assert_array_almost_equal(state['x'], [2.5, 7, 4.5])
    npt.assert_array_almost_equal(state['y'], xy3tb.optimal[:, 1])


def test_TopFarmProblemSpacingPenalty():
    tf = xy3tb.get_tf(design_vars={'x': [3, 7, 4], 'y': [-3, -7, -3]},
                      driver=EasyRandomSearchDriver(RandomizeTurbinePosition(1), 10),
                      constraints=[SpacingConstraint(2)])
    # spacing violated
    cost, _ = tf.evaluate({'x': xy3tb.desired[:, 0], 'y': xy3tb.desired[:, 1]})
    npt.assert_array_less(1e10, cost)

    # spacing satisfied
    cost, _ = tf.evaluate({'x': xy3tb.optimal[:, 0], 'y': xy3tb.optimal[:, 1]})
    npt.assert_equal(1.5, cost)


def test_TopFarmProblemXYBoundaryConstraint():
    tf = xy3tb.get_tf(design_vars={'x': [3, 7, 4], 'y': [-3, -7, -3]},
                      constraints=[XYBoundaryConstraint(xy3tb.boundary)])
    tf.evaluate({'x': xy3tb.desired[:, 0], 'y': xy3tb.desired[:, 1]})
    npt.assert_equal(tf['boundaryDistances'][1, 3], -1)

    _, state, _ = tf.optimize()
    npt.assert_array_almost_equal(state['x'], [3, 6, 4])
    npt.assert_array_almost_equal(state['y'], xy3tb.optimal[:, 1])

    desvars = tf.driver._designvars
    for xy in 'xy':
        for lu in ['lower', 'upper']:
            npt.assert_equal(desvars['indeps.' + xy][lu], np.nan)


def test_TopFarmProblemXYBoundaryConstraintPolygon():
    tf = xy3tb.get_tf(design_vars={'x': [3, 7, 4], 'y': [-3, -7, -3]},
                      constraints=[XYBoundaryConstraint(xy3tb.boundary, 'polygon')])
    # constraint violated
    tf.evaluate({'x': xy3tb.desired[:, 0], 'y': xy3tb.desired[:, 1]})
    npt.assert_equal(tf['boundaryDistances'][1], -1)

    _, state, _ = tf.optimize()
    npt.assert_array_almost_equal(state['x'], [3, 6, 4])
    npt.assert_array_almost_equal(state['y'], xy3tb.optimal[:, 1])


def test_TopFarmProblemXYBoundaryPenalty():
    tf = xy3tb.get_tf(design_vars={'x': [3, 7, 4], 'y': [-3, -7, -3]},
                      driver=EasyRandomSearchDriver(RandomizeTurbinePosition(1), 10),
                      constraints=[XYBoundaryConstraint(xy3tb.boundary)])
    # spacing violated
    cost, _ = tf.evaluate({'x': xy3tb.desired[:, 0], 'y': xy3tb.desired[:, 1]})
    npt.assert_array_less(1e10, cost)

    # spacing satisfied
    cost, _ = tf.evaluate({'x': xy3tb.optimal[:, 0], 'y': xy3tb.optimal[:, 1]})
    npt.assert_equal(1.5, cost)


def test_TopFarmProblemXYBoundaryPenaltyAndLimits():
    tf = xy3tb.get_tf(design_vars={'x': ([3, 7, 4], -1, 5), 'y': ([-3, -7, -3], -9, -1)},
                      driver=EasyRandomSearchDriver(RandomizeTurbinePosition(1), 10),
                      constraints=[XYBoundaryConstraint(xy3tb.boundary)])
    tf.evaluate({'x': xy3tb.desired[:, 0], 'y': xy3tb.desired[:, 1]})
    npt.assert_equal(tf['boundaryDistances'][1, 3], -1)

    desvars = tf.driver._designvars
    npt.assert_equal(desvars['indeps.x']['lower'], 0)
    npt.assert_equal(desvars['indeps.x']['upper'], 5)
    npt.assert_array_equal(desvars['indeps.y']['lower'], -9)
    npt.assert_array_equal(desvars['indeps.y']['upper'], -1)



"""
2018 - Created by Witold Skrzypinski, wisk@dtu.dk
Implementation of Offshore Turbine Cost Model, ref. Soeren Oemann Lind, soli@dtu.dk
Structure based on 2017 turbine_cost.py by Arvydas Berzonskis, Goldwind
"""
import numpy as np


class economic_evaluation():
    '''All masses are in kg, all costs are in 2017 Euro'''

    def __init__(self, distance_from_shore, energy_price, project_duration,
                 discount_rate=0.0909870634):

        # Initialize dictionaries
        self.project_costs = {}
        self.project_costs_sums = {}
        self.discount_rate = discount_rate              # [-]
        self.distance_from_shore = distance_from_shore  # [km]
        self.energy_price = energy_price                # [Euro/kWh]
        self.project_duration = project_duration        # [years]

    def calculate_npv(self, rated_rpm_array, D_rotor_array, Power_rated_array,
                      hub_height_array, water_depth_array, aep_array, cabling_cost=None):
        '''
        Calculate Net Present Value [Euro]

        Input:
        rated_rpm_array   [RPM]
        D_rotor_array     [m]
        Power_rated_array [MW]
        hub_height_array  [m]
        aep_array         [kWh]
        water_depth_array [m]
        '''

        self.calculate_expenditures(
            rated_rpm_array,
            D_rotor_array,
            Power_rated_array,
            hub_height_array,
            aep_array,
            water_depth_array,
            cabling_cost)

        self.calculate_cash_flow()

        self.NPV = np.npv(self.discount_rate, self.CWF)

        return self.NPV

    def calculate_irr(self, rated_rpm_array, D_rotor_array, Power_rated_array,
                      hub_height_array, water_depth_array, aep_array, cabling_cost=None):
        '''
        Calculate Internal Rate of Return [%]

        Input:
        rated_rpm_array   [RPM]
        D_rotor_array     [m]
        Power_rated_array [MW]
        hub_height_array  [m]
        aep_array         [kWh]
        water_depth_array [m]
        '''

        self.calculate_expenditures(
            rated_rpm_array,
            D_rotor_array,
            Power_rated_array,
            hub_height_array,
            aep_array,
            water_depth_array,
            cabling_cost)

        self.calculate_cash_flow()

        self.IRR = 100 * np.irr(self.CWF)  # [%]

        return self.IRR

    def calculate_cash_flow(self):
        '''Calculate cash flow every year during the project period'''

        self.CWF = []

        for i in range(self.project_duration + 1):

            if i == 0:
                self.CWF.append(-self.project_costs_sums["DEVEX"] - self.project_costs_sums["CAPEX"])

            elif i == self.project_duration:
                self.CWF.append(int(self.energy_price * sum(self.aep_vector) -
                                    self.project_costs_sums["ABEX"] -
                                    self.project_costs_sums["OPEX"]))

            else:
                self.CWF.append(int(self.energy_price * sum(self.aep_vector) -
                                    self.project_costs_sums["OPEX"]))

    def calculate_expenditures(self, rated_rpm, rotor_diameter, rated_power,
                               hub_height, aep_array, water_depth, cabling_cost=None):
        '''Calculate DEVEX, CAPEX, OPEX and ABEX [Euro]'''

        # Ensure that the variables are numpy arrays
        self.aep_vector = np.array(aep_array)

        # Calculate expenditures
        self.calculate_devex(rated_power)

        self.calculate_capex(rated_rpm, rotor_diameter, rated_power,
                             hub_height, water_depth, cabling_cost)

        self.calculate_opex(rated_power)

        self.calculate_abex()

    def calculate_devex(self, rated_power):
        '''Calculate DEVEX [Euro]'''

        rated_power = np.array(rated_power)
        number_of_turbines = np.size(rated_power)

        myName = "DEVEX"
        magicFactor = 2.0e3 / 39 * 0.3 / 80

        self.project_costs[myName] = {
            "environmental_survey": magicFactor * 1e6 * self.distance_from_shore,
            "sea_bed_survey": 1.5 * 1e5 * number_of_turbines,
            "met_mast": magicFactor * 1e6 * self.distance_from_shore,
            "development_services": 0.7 * 1e5 * np.sum(rated_power)}

        self.project_costs_sums[myName] = sum(
            self.project_costs[myName].values())

    def calculate_capex(self, rated_rpm, rotor_diameter, rated_power,
                        hub_height, water_depth, cabling_cost=None):
        '''Calculate CAPEX [Euro]'''

        rated_rpm = np.array(rated_rpm)
        rotor_diameter = np.array(rotor_diameter)
        rated_power = np.array(rated_power)
        hub_height = np.array(hub_height)
        water_depth = np.array(water_depth)

        self.calculate_turbine(
            rated_rpm,
            rotor_diameter,
            rated_power,
            hub_height)

        self.calculate_foundation(rotor_diameter, rated_power, water_depth)

        # %% BOP  [Euro] (Balance of Plant costs)

        myName = "BOP"

        self.project_costs[myName] = {
            "substation": 6.0e6 +
            1.0e4 *
            np.sum(rated_power) ** 1.5,
            "array_of_cables": 3.5e3 *
            np.sum(rotor_diameter),
            "cables_export": 1.4e6 *
            self.distance_from_shore,
            "onshore_electrical": 5.0e4 *
            np.sum(rated_power) +
            50.0 *
            np.sum(rated_power) ** 2.0}

        if cabling_cost:
            self.project_costs[myName]["array_of_cables"] = cabling_cost

        self.project_costs_sums[myName] = sum(
            self.project_costs[myName].values())

        # %% CAPEX [Euro]

        self.project_costs_sums["CAPEX"] = sum(
            self.turbine_general_costs_sums["bill_of_material"]) + sum(
            self.foundation_general_costs_sums["bill_of_material"]) + self.project_costs_sums["BOP"]

    def calculate_opex(self, rated_power):
        ''' Calculate OPEX [Euro / year]'''

        rated_power = np.array(rated_power)
        number_of_turbines = np.size(rated_power)

        myName = "O&M"

        self.project_costs[myName] = {
            "onshore personnel": 5.5e3 *
            np.sum(rated_power),
            "buildings, habor fees etc.": 3.0e6,
            "mobilization, rental time": 9.6e4 *
            number_of_turbines,
            "jackup personnel": 8.5e3 *
            number_of_turbines,
            "offshore service personnel": 1.2e4 *
            np.sum(rated_power),
            "service, failed components": 2.1e3 *
            np.sum(rated_power),
            "ships, offhsore operations": 4.6e3 *
            number_of_turbines *
            self.distance_from_shore}

        self.project_costs_sums[myName] = sum(
            self.project_costs[myName].values())

        self.project_costs_sums["OPEX"] = self.project_costs_sums["O&M"]

    def calculate_abex(self):
        ''' Calculate ABEX [Euro]'''

        self.project_costs_sums["ABEX"] = 0.03 * \
            self.project_costs_sums["CAPEX"]

    def calculate_turbine(
            self,
            rated_rpm,
            rotor_diameter,
            rated_power,
            hub_height):
        ''' Calculate the part of CAPEX associated with the turbine itself [Euro]'''

        self.turbine_component_mass = {}
        self.turbine_component_mass_sums = {}
        self.turbine_component_costs = {}
        self.turbine_component_costs_sums = {}
        self.turbine_general_costs = {}
        self.turbine_general_costs_sums = {}

        rated_torque = rated_power / (rated_rpm * np.pi / 30.0) * 1.1
        rotor_area = np.pi * (rotor_diameter / 2) ** 2

        # %% Blades

        comp_name = "3_blades"

        self.turbine_component_mass[comp_name] = 0.0 + \
            0.3 * rotor_diameter ** 2.5
        self.turbine_component_costs[comp_name] = 12.0 * \
            self.turbine_component_mass[comp_name]

        self.turbine_component_mass_sums[comp_name] = self.turbine_component_mass[comp_name]
        self.turbine_component_costs_sums[comp_name] = self.turbine_component_costs[comp_name]

        # %% Hub incl. pitch system w/ bearings

        comp_name = "hub"

        self.turbine_component_mass[comp_name] = {
            "structure": 6.0e3 + 0.1 * rotor_diameter ** 2.5,
            "pitch_bearings": 5.0e2 + 0.07 * rotor_diameter ** 2.5,
            "pitch_system": 5.0e2 + 0.03 * rotor_diameter ** 2.5,
            "secondary": 7.0e2 + 15.0 * rotor_diameter ** 1.0}

        self.turbine_component_costs[comp_name] = {
            "structure": 2.5 * self.turbine_component_mass[comp_name]["structure"],
            "pitch_bearings": 8.0 * self.turbine_component_mass[comp_name]["pitch_bearings"],
            "pitch_system": 8.0 * self.turbine_component_mass[comp_name]["pitch_system"],
            "secondary": 8.0 * self.turbine_component_mass[comp_name]["secondary"]}

        self.turbine_component_mass_sums[comp_name] = sum(
            self.turbine_component_mass[comp_name].values())
        self.turbine_component_costs_sums[comp_name] = sum(
            self.turbine_component_costs[comp_name].values())

        # %% High Speed Drivetrain
        # self.high_speed_drivetrain(rotor_diameter, rated_torque, rated_rpm)

        # %% Medium Speed Drivetrain
        self.medium_speed_drivetrain(rotor_diameter, rated_torque, rated_rpm)

        # %% Direct Drive Drivetrain
        # self.direct_drive_drivetrain(rotor_diameter, rated_torque)

        # %% Nacelle

        comp_name = "nacelle"

        self.turbine_component_mass[comp_name] = {
            "cooling": 0.0 + 500.0 * rated_power ** 1.0,
            "converter": 0.0 + 1.0e3 * rated_power ** 1.0,
            "controller": 200.0 + 100.0 * rated_power ** 1.0,
            "yaw": 0.0 + 0.1 * rotor_diameter ** 2.5,
            "canopy": 1.0e3 + 1.5e3 * rated_power ** 1.0,
            "secondary": 1.0e3 + 1.0e3 * rated_power ** 1.0}

        self.turbine_component_costs[comp_name] = {
            "cooling": 8.0 * self.turbine_component_mass[comp_name]["cooling"],
            "converter": 30.0 * self.turbine_component_mass[comp_name]["converter"],
            "controller": 50.0 * self.turbine_component_mass[comp_name]["controller"],
            "yaw": 6.0 * self.turbine_component_mass[comp_name]["yaw"],
            "canopy": 10.0 * self.turbine_component_mass[comp_name]["canopy"],
            "secondary": 10.0 * self.turbine_component_mass[comp_name]["secondary"]}

        self.turbine_component_mass_sums[comp_name] = sum(
            self.turbine_component_mass[comp_name].values())
        self.turbine_component_costs_sums[comp_name] = sum(
            self.turbine_component_costs[comp_name].values())

        # %% Tower

        comp_name = "tower"

        self.turbine_component_mass[comp_name] = {
            "structure": 0.0 + 0.25 * (rotor_area * hub_height) ** 1.0,
            "internal": 1.0e3 + 100.0 * hub_height ** 1.0,
            "cabling": 0.0 + 25.0 * (rated_power * hub_height) ** 1.0,
            "secondary": 1.0e3 + 500.0 * rated_power ** 1.0,
            "transformer": 0.0 + 2.5e3 * rated_power ** 1}

        self.turbine_component_costs[comp_name] = {
            "structure": 1.5 * self.turbine_component_mass[comp_name]["structure"],
            "internal": 8.0 * self.turbine_component_mass[comp_name]["internal"],
            "cabling": 8.0 * self.turbine_component_mass[comp_name]["cabling"],
            "secondary": 10.0 * self.turbine_component_mass[comp_name]["secondary"],
            "transformer": 8.0 * self.turbine_component_mass[comp_name]["transformer"]}

        self.turbine_component_mass_sums[comp_name] = sum(
            self.turbine_component_mass[comp_name].values())
        self.turbine_component_costs_sums[comp_name] = sum(
            self.turbine_component_costs[comp_name].values())

        # %% Turbine Bill of Material (BOM) cost

        self.turbine_general_costs_sums["bill_of_material"] = sum(
            self.turbine_component_costs_sums.values())

        # %% Turbine Direct Production Cost

        myName = "direct_production"

        self.turbine_general_costs[myName] = {
            "direct_labor": 0.03 *
            self.turbine_general_costs_sums["bill_of_material"],
            "production_overhead": 0.1 *
            self.turbine_general_costs_sums["bill_of_material"]}

        self.turbine_general_costs_sums[myName] = sum(self.turbine_general_costs[myName].values(
        )) + self.turbine_general_costs_sums["bill_of_material"]

        # %% Turbine Selling, General and Administrative Expenses (SG&A)

        myName = "SG&A"

        self.turbine_general_costs[myName] = {
            "SG&A_overhead": 0.05 * (self.turbine_general_costs_sums["direct_production"]),
            "R&D": 0.03 * (self.turbine_general_costs_sums["direct_production"]),
            "SG&A": 0.05 * (self.turbine_general_costs_sums["direct_production"])}

        self.turbine_general_costs_sums[myName] = sum(
            self.turbine_general_costs[myName].values())

        # %% Turbine Total Production Cost

        self.turbine_general_costs_sums["total_production"] = self.turbine_general_costs_sums[
            "direct_production"] + self.turbine_general_costs_sums["SG&A"]

        # %% Turbine Project costs

        myName = "project"

        self.turbine_general_costs[myName] = {
            "transportation": 0.2 * sum(self.turbine_component_mass_sums.values()) + 1.0e4,
            "harbor_storage_and_assembly": 2.5e4 * rated_power + 1.5e5,
            "installation_and_comissioning": 5.0e4 * rated_power + 1.0e5,
            "warranty_and_accruals": 0.03 * self.turbine_general_costs_sums["total_production"],
            "financing": 0.02 * self.turbine_general_costs_sums["total_production"]}

        self.turbine_general_costs_sums[myName] = sum(
            self.turbine_general_costs[myName].values())

        # %% Turbine Total Cost

        self.turbine_general_costs_sums["TOTAL"] = self.turbine_general_costs_sums["total_production"] + \
            self.turbine_general_costs_sums["project"]

    def calculate_foundation(self, rotor_diameter, rated_power, water_depth):
        ''' Calculate the part of CAPEX associated with the foundation itself [Euro]'''

        # Initialize dictionaries
        self.foundation_mass = {}
        self.foundation_costs = {}
        self.foundation_general_costs = {}
        self.foundation_general_costs_sums = {}

        # %% Foundation

        self.foundation_mass = {
            "monopile_incl_TP": (
                6.5e4 +
                4.5e3 *
                water_depth +
                40.0 *
                water_depth**2) *
            rated_power}

        self.foundation_costs = {
            "monopile_incl_TP": 1.5 * self.foundation_mass["monopile_incl_TP"]}
        '''
        self.foundation_mass = {
        "jacket_incl_boat_landing" : ( 8.3e4 - 700.0 * water_depth + 110.0 * water_depth**2) * rated_power}

        self.foundation_costs = {
        "jacket_incl_boat_landing" : 2.4 * self.foundation_mass["jacket_incl_boat_landing"]}
        '''
        self.foundation_mass_sum = sum(self.foundation_mass.values())
        self.foundation_cost_sum = sum(self.foundation_costs.values())

        # %% Foundation Bill of Material (BOM) cost

        self.foundation_general_costs_sums["bill_of_material"] = self.foundation_cost_sum

        # %% Foundation Direct Production Cost

        myName = "direct_production"

        self.foundation_general_costs[myName] = {
            "direct_labor": 0.03 *
            self.foundation_general_costs_sums["bill_of_material"],
            "material_overhead": 0.1 *
            self.foundation_general_costs_sums["bill_of_material"]}

        self.foundation_general_costs_sums[myName] = sum(self.foundation_general_costs[myName].values(
        )) + self.foundation_general_costs_sums["bill_of_material"]

        # %% Foundation Selling, General and Administrative Expenses (SG&A)

        myName = "SG&A"

        self.foundation_general_costs[myName] = {
            "SG&A_overhead": 0.05 * (self.foundation_general_costs_sums["direct_production"]),
            "R&D": 0.03 * (self.foundation_general_costs_sums["direct_production"]),
            "SG&A": 0.05 * (self.foundation_general_costs_sums["direct_production"])}

        self.foundation_general_costs_sums[myName] = sum(
            self.foundation_general_costs[myName].values())

        # %% Foundation Total Production Cost

        self.foundation_general_costs_sums["total_production"] = self.foundation_general_costs_sums[
            "direct_production"] + self.foundation_general_costs_sums["SG&A"]

        # %% Foundation Project costs

        myName = "project"

        self.foundation_general_costs[myName] = {
            "transportation": 0.15 *
            self.foundation_general_costs_sums["total_production"],
            "installation": 160.0 *
            self.foundation_mass_sum ** 0.5 +
            2.0e3 *
            self.distance_from_shore,
            "warranty_and_accruals": 0.03 *
            self.foundation_general_costs_sums["total_production"],
            "financing": 0.02 *
            self.foundation_general_costs_sums["total_production"]}

        self.foundation_general_costs_sums[myName] = sum(
            self.foundation_general_costs[myName].values())

        # %% Foundation Total Cost

        self.foundation_general_costs_sums["TOTAL"] = self.foundation_general_costs_sums[
            "total_production"] + self.foundation_general_costs_sums["project"]

    def high_speed_drivetrain(self, rotor_diameter, rated_torque, rated_rpm):
        '''Calculate the cost of high speed drivetrain [Euro]'''

        comp_name = "hs_drivetrain"

        self.turbine_component_mass[comp_name] = {
            "bedplate": 0.0 + 2.4 * rotor_diameter ** 2.0,
            "main_shaft": 0.0 + 0.02 * rotor_diameter ** 2.8,
            "main_bearings": 0.0 + 0.02 * rotor_diameter ** 2.5,
            "bearing_housing": 0.0 + 0.03 * rotor_diameter ** 2.5,
            "gearbox": 0.0 + 15.0e3 * rated_torque ** 1.0,
            "coupling_&_brake": 500.0 + 50.0 * (rated_torque * rated_rpm / 1.5) ** 1.0,
            "generator": 1.0e3 + 400.0 * (rated_torque * rated_rpm / 1.5) ** 1.0}

        self.turbine_component_costs[comp_name] = {
            "bedplate": 2.5 *
            self.turbine_component_mass[comp_name]["bedplate"],
            "main_shaft": 5.0 *
            self.turbine_component_mass[comp_name]["main_shaft"],
            "main_bearings": 15.0 *
            self.turbine_component_mass[comp_name]["main_bearings"],
            "bearing_housing": 2.5 *
            self.turbine_component_mass[comp_name]["bearing_housing"],
            "gearbox": 8.0 *
            self.turbine_component_mass[comp_name]["gearbox"],
            "coupling_&_brake": 8.0 *
            self.turbine_component_mass[comp_name]["coupling_&_brake"],
            "generator": 8.0 *
            self.turbine_component_mass[comp_name]["generator"]}

        self.turbine_component_mass_sums[comp_name] = sum(
            self.turbine_component_mass[comp_name].values())
        self.turbine_component_costs_sums[comp_name] = sum(
            self.turbine_component_costs[comp_name].values())

    def medium_speed_drivetrain(self, rotor_diameter, rated_torque, rated_rpm):
        '''Calculate the cost of medium speed drivetrain [Euro]'''

        comp_name = "ms_drivetrain"

        self.turbine_component_mass[comp_name] = {
            "bedplate": 0.0 + 1.2 * rotor_diameter ** 2.0,
            "main_shaft": 0.0 + 0.02 * rotor_diameter ** 2.8,
            "main_bearings": 0.0 + 0.02 * rotor_diameter ** 2.5,
            "bearing_housing": 0.0 + 0.03 * rotor_diameter ** 2.5,
            "gearbox": 0.0 + 1.0e4 * rated_torque ** 1.0,
            "coupling_&_brake": 0.0 + 30.0 * (rated_torque * rated_rpm / 0.4) ** 1.0,
            "generator": 6.0e3 + 100.0 * (rated_torque * rated_rpm / 0.4) ** 1.0}

        self.turbine_component_costs[comp_name] = {
            "bedplate": 2.5 *
            self.turbine_component_mass[comp_name]["bedplate"],
            "main_shaft": 5.0 *
            self.turbine_component_mass[comp_name]["main_shaft"],
            "main_bearings": 15.0 *
            self.turbine_component_mass[comp_name]["main_bearings"],
            "bearing_housing": 2.5 *
            self.turbine_component_mass[comp_name]["bearing_housing"],
            "gearbox": 8.0 *
            self.turbine_component_mass[comp_name]["gearbox"],
            "coupling_&_brake": 8.0 *
            self.turbine_component_mass[comp_name]["coupling_&_brake"],
            "generator": 8.0 *
            self.turbine_component_mass[comp_name]["generator"]}

        self.turbine_component_mass_sums[comp_name] = sum(
            self.turbine_component_mass[comp_name].values())
        self.turbine_component_costs_sums[comp_name] = sum(
            self.turbine_component_costs[comp_name].values())

    def direct_drive_drivetrain(self, rotor_diameter, rated_torque):
        '''Calculate the cost of direct drive drivetrain [Euro]'''

        comp_name = "dd_drivetrain"

        self.turbine_component_mass[comp_name] = {
            "bedplate": 0.0 + 1.2 * rotor_diameter ** 2.0,
            "generator": 0.0 + 35.0e3 * rated_torque ** 0.666666666667}

        self.turbine_component_costs[comp_name] = {
            "bedplate": 2.5 * self.turbine_component_mass[comp_name]["bedplate"],
            "generator": 8.0 * self.turbine_component_mass[comp_name]["generator"]}

        self.turbine_component_mass_sums[comp_name] = sum(
            self.turbine_component_mass[comp_name].values())
        self.turbine_component_costs_sums[comp_name] = sum(
            self.turbine_component_costs[comp_name].values())



import warnings
import numpy as np
from numpy.random import RandomState
import pandas as pd
from sklearn.preprocessing import StandardScaler, MinMaxScaler
from sklearn.neural_network import MLPRegressor, MLPClassifier

from topfarm.constraint_components.load import (
    SurrogateModel,
    compute_error,
    predict_output,
    predict_gradient)


def test_SurrogateModel():
    """
    Test SurrogateModel class.
    """

    obj = SurrogateModel(model=MLPRegressor(),
                         input_scaler=MinMaxScaler(),
                         output_scaler=StandardScaler(),
                         input_channel_names=['x0', 'x1'],
                         output_channel_name='y')

    assert obj is not None


def test_compute_error():
    """
    Basic test of compute_error().
    """

    data_1 = np.array([2.0, 2.0])
    data_2 = np.array([1.0, 3.0])

    abs_err, rel_err = compute_error(data_1, data_2)

    assert np.array_equal(abs_err, np.array([-1.0, 1.0]))
    assert np.array_equal(rel_err, np.array([-0.5, 0.5]))


def test_predict_output_1():
    """
    Basic test of predict_output().
    """

    prng = RandomState(seed=0)
    input = prng.rand(5, 2)

    def model(input):
        return (input[:, 0] * input[:, 1]).reshape(-1, 1)

    output_ok = np.array(model(input))
    output_try, _ = predict_output(model, input)

    assert np.array_equal(output_ok, output_try)


def test_predict_output_2():
    """
    Test of predict_output() when input is a dict.
    """

    input = {'x0': [1.0, 2.0, 3.0],
             'x1': [4.0, 5.0, 6.0]}

    def model(input):
        return (input[:, 0] * input[:, 1]).reshape(-1, 1)

    output_ok = np.array(model(np.array([input['x0'], input['x1']]).transpose()))
    output_try, _ = predict_output(model, input, model_in_keys=list(input))

    assert np.array_equal(output_ok, output_try)


def test_predict_output_with_input_scaler():
    """
    Test with an input scaler.
    """

    prng = RandomState(seed=0)
    input = -100.0 + 200.0 * prng.rand(5, 2)
    input_scaler = MinMaxScaler()
    input_scaled = input_scaler.fit_transform(input)

    def model(input):
        return (input[:, 0] * input[:, 1]).reshape(-1, 1)

    output_ok = model(input_scaled)
    output_try, _ = predict_output(model, input, input_scaler=input_scaler)

    assert np.array_equal(output_ok, output_try)


def test_predict_output_with_output_scaler():
    """
    Test with an output scaler.
    """

    prng = RandomState(seed=0)
    input = prng.rand(5, 2)

    def model(input):
        return (- 100.0 + 200.0 * input[:, 0] * input[:, 1]).reshape(-1, 1)

    output_ok = model(input)
    output_scaler_mm = MinMaxScaler()
    output_scaler_std = StandardScaler()
    output_scaler_mm.fit(output_ok)
    output_scaler_std.fit(output_ok)

    def model_scaled_mm(input):
        output = model(input)
        return output_scaler_mm.transform(output)

    def model_scaled_std(input):
        output = model(input)
        return output_scaler_std.transform(output)

    output_try_mm, _ = predict_output(model_scaled_mm, input,
                                      output_scaler=output_scaler_mm)
    output_try_std, _ = predict_output(model_scaled_std, input,
                                       output_scaler=output_scaler_std)
    output_try_mm = output_try_mm
    output_try_std = output_try_std

    assert np.allclose(output_ok, output_try_mm)
    assert np.allclose(output_ok, output_try_std)


def test_wrong_input_type():
    """
    Test of predict_output() and predict_gradient() when input has the wrong
    type.
    """

    input = pd.DataFrame()

    def model():
        pass

    try:
        predict_output(model, input)
        assert False
    except TypeError:
        assert True

    try:
        predict_gradient(model, input)
        assert False
    except TypeError:
        assert True


def test_boundary():
    """
    Test of predict_output() when there is a boundary.
    """

    prng = RandomState(seed=0)
    input = -1 + 2 * prng.rand(10, 1)
    input[2] = 10
    input[4] = -5

    def model(input):
        return 2 * input

    def bound_fun(x):
        return np.abs(x) <= 1

    with warnings.catch_warnings(record=True) as w:
        _, extrapolation_sample = predict_output(
            model, input, boundary=bound_fun)
        assert np.array_equal(extrapolation_sample, [2, 4])
        assert 'evaluated outside' in str(w[-1].message)

    with warnings.catch_warnings(record=True) as w:
        _, extrapolation_sample = predict_gradient(
            model, input, boundary=bound_fun)
        assert np.array_equal(extrapolation_sample, [2, 4])
        assert 'evaluated outside' in str(w[-1].message)


def test_wrong_model_type():
    """
    Test of predict_output() and predict_gradient() when the model type is
    unknown.
    """

    input = np.random.rand(5, 1)
    input_scaler = MinMaxScaler()
    input_scaler.fit(input)

    model = MLPClassifier()

    try:
        predict_output(model, input)
        assert False
    except KeyError:
        assert True

    try:
        predict_gradient(model, input)
        assert False
    except KeyError:
        assert True

    try:
        predict_gradient(model, input, input_scaler=input_scaler)
        assert False
    except KeyError:
        assert True


def test_predict_gradient_1():
    """
    Basic test of predict_gradient
    """

    input = {'x0': [1.0, 2.0, 3.0],
             'x1': [4.0, 5.0, 6.0]}

    def fun(input):
        x0, x1 = input.T
        return x0 ** 2 * np.sin(x1)

    def Dfun_Dx0(input):
        x0, x1 = input.T
        return 2 * x0 * np.sin(x1)

    def Dfun_Dx1(input):
        x0, x1 = input.T
        return x0 ** 2 * np.cos(x1)

    def Dfun(input):
        x0, x1 = input.T
        return np.column_stack((2 * x0 * np.sin(x1), x0 ** 2 * np.cos(x1)))

    grad_x0, _ = predict_output(Dfun_Dx0, input, model_in_keys=list(input))
    grad_x1, _ = predict_output(Dfun_Dx1, input, model_in_keys=list(input))
    grad_ok = np.column_stack((grad_x0, grad_x1))

    grad_test, _ = predict_gradient(Dfun, input, model_in_keys=['x0', 'x1'])

    assert np.array_equal(grad_ok, grad_test)


def test_predict_gradient_with_input_scaler():
    """
    Test predict_gradient with an input scaler.
    """

    prng = RandomState(seed=0)
    input = -100.0 + 200.0 * prng.rand(5, 2)
    input_scaler_mm = MinMaxScaler()
    input_scaled_mm = input_scaler_mm.fit_transform(input)
    input_scaler_std = StandardScaler()
    input_scaled_std = input_scaler_std.fit_transform(input)

    def fun(input):
        x0, x1 = input.T
        return x0 ** 2 * np.sin(x1)

    def Dfun_Dx0(input):
        x0, x1 = input.T
        return 2 * x0 * np.sin(x1)

    def Dfun_Dx1(input):
        x0, x1 = input.T
        return x0 ** 2 * np.cos(x1)

    def Dfun(input):
        x0, x1 = input.T
        return np.column_stack((2 * x0 * np.sin(x1), x0 ** 2 * np.cos(x1)))

    # Test input_scaler_mm
    grad_x0, _ = predict_output(Dfun_Dx0, input_scaled_mm)
    grad_x1, _ = predict_output(Dfun_Dx1, input_scaled_mm)
    grad_ok = np.column_stack((grad_x0, grad_x1)) * input_scaler_mm.scale_

    grad_test, _ = predict_gradient(Dfun, input, input_scaler=input_scaler_mm)

    assert np.array_equal(grad_ok, grad_test)

    # Test input_scaler_std
    grad_x0, _ = predict_output(Dfun_Dx0, input_scaled_std)
    grad_x1, _ = predict_output(Dfun_Dx1, input_scaled_std)
    grad_ok = np.column_stack((grad_x0, grad_x1)) / input_scaler_std.scale_

    grad_test, _ = predict_gradient(Dfun, input, input_scaler=input_scaler_std)

    assert np.array_equal(grad_ok, grad_test)


def test_predict_gradient_with_output_scaler():
    """
    Test predict_gradient with an output scaler.
    """

    prng = RandomState(seed=0)
    input = prng.rand(5, 2)

    def fun(input):
        x0, x1 = input.T
        return (x0 ** 2 * np.sin(x1)).reshape(-1, 1)

    output_ok = fun(input)
    output_scaler_mm = MinMaxScaler()
    output_scaler_std = StandardScaler()
    output_scaler_mm.fit(output_ok)
    output_scaler_std.fit(output_ok)

    def Dfun_Dx0(input):
        x0, x1 = input.T
        return 2 * x0 * np.sin(x1)

    def Dfun_Dx1(input):
        x0, x1 = input.T
        return x0 ** 2 * np.cos(x1)

    def Dfun(input):
        x0, x1 = input.T
        return np.column_stack((2 * x0 * np.sin(x1), x0 ** 2 * np.cos(x1)))

    # Test input_scaler_mm
    grad_x0, _ = predict_output(Dfun_Dx0, input)
    grad_x1, _ = predict_output(Dfun_Dx1, input)
    grad_ok = np.column_stack((grad_x0, grad_x1)) / output_scaler_mm.scale_

    grad_test, _ = predict_gradient(Dfun, input, output_scaler=output_scaler_mm)

    assert np.array_equal(grad_ok, grad_test)

    # Test input_scaler_std
    grad_x0, _ = predict_output(Dfun_Dx0, input)
    grad_x1, _ = predict_output(Dfun_Dx1, input)
    grad_ok = np.column_stack((grad_x0, grad_x1)) * output_scaler_std.scale_

    grad_test, _ = predict_gradient(Dfun, input, output_scaler=output_scaler_std)

    assert np.array_equal(grad_ok, grad_test)


def test_MLPRegressor():
    """
    Test predict_output() when model is a MLPRegressor.
    """

    prng = RandomState(seed=0)
    input = prng.rand(100, 1)
    output = 2.0 * input
    model = MLPRegressor(
        hidden_layer_sizes=(1),
        solver='lbfgs',
        alpha=1e-4,
        max_iter=3000,
        activation='identity',
        learning_rate='adaptive')
    model.fit(input, output.ravel())

    predicted_output, _ = predict_output(model, np.array(input))

    assert np.allclose(output, predicted_output, atol=1.e-3)

#    import matplotlib.pyplot as plt
#    fig, ax = plt.subplots()
#    ax.scatter(input, output)
#    ax.scatter(input, predicted_output)



from openmdao.drivers.doe_driver import DOEDriver
from openmdao.drivers.doe_generators import FullFactorialGenerator
import pytest
import numpy as np
from topfarm import ProblemComponent
from topfarm import TopFarmProblem
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm.cost_models.dummy import DummyCost
from topfarm.tests import npt
from topfarm.constraint_components.spacing import SpacingConstraint
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.easy_drivers import EasyScipyOptimizeDriver
from topfarm.tests.test_files import xy3tb

"""Test methods in TopFarmProblem
cost
state
state_array
update state
evaluate
optimize
check_gradients
as_component
get_DOE_list
get_DOE_array
turbine_positions
smart_start
"""


@pytest.fixture
def turbineTypeOptimizationProblem():
    return TopFarmProblem(
        design_vars={'type': ([0, 0, 0], 0, 2)},
        cost_comp=DummyCost(np.array([[2, 0, 1]]).T, ['type']),
        driver=DOEDriver(FullFactorialGenerator(3)))


@pytest.mark.parametrize('design_vars', [{'type': ([0, 0, 0], 0, 2)},
                                         [('type', ([0, 0, 0], 0, 2))],
                                         (('type', ([0, 0, 0], 0, 2)),),
                                         zip(['type'], [([0, 0, 0], 0, 2)]),
                                         ])
def test_design_var_list(turbineTypeOptimizationProblem, design_vars):
    tf = TopFarmProblem(
        design_vars=design_vars,
        cost_comp=DummyCost(np.array([[2, 0, 1]]).T, ['type']),
        driver=DOEDriver(FullFactorialGenerator(3)))
    cost, _, = tf.evaluate()
    npt.assert_equal(tf.cost, cost)
    assert tf.cost == 5


def test_cost(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    cost, _, = tf.evaluate()
    npt.assert_equal(tf.cost, cost)
    assert tf.cost == 5


def test_state(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    npt.assert_equal(tf.state, {'type': [0, 0, 0]})


def test_state_array(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    arr = tf.state_array(['type', 'type'])
    npt.assert_equal(arr.shape, [3, 2])
    npt.assert_array_equal(arr, [[0, 0],
                                 [0, 0],
                                 [0, 0]])


@pytest.mark.parametrize('types,cost', [([0, 0, 0], 5),
                                        ([2, 0, 2], 1)])
def test_update_state(turbineTypeOptimizationProblem, types, cost):
    tf = turbineTypeOptimizationProblem
    c, state = tf.evaluate({'type': types})
    npt.assert_equal(c, cost)
    npt.assert_array_equal(state['type'], types)
    # wrong shape
    c, state = tf.evaluate({'type': [types]})
    npt.assert_equal(c, cost)
    npt.assert_array_equal(state['type'], types)
    # missing key
    c, state = tf.evaluate({'missing': types})


def test_evaluate(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    cost, state = tf.evaluate()
    assert cost == 5
    np.testing.assert_array_equal(state['type'], [0, 0, 0])
    tf.evaluate(disp=True)  # test that disp=True does not fail


def test_optimize(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    cost, state, recorder = tf.optimize()
    assert cost == 0
    np.testing.assert_array_equal(state['type'], [2, 0, 1])
    doe_list = np.squeeze(tf.get_DOE_array())
    np.testing.assert_array_almost_equal(recorder.get('cost'), np.sum((doe_list - [2, 0, 1])**2, 1))
    tf.optimize(disp=True)  # test that disp=True does not fail


initial = np.array([[6, 0, 70, 0],
                    [6, -8, 71, 1],
                    [1, 1, 72, 2],
                    [-1, -8, 73, 3]])  # initial turbine layouts
optimal = np.array([[3, -3], [7, -7], [4, -3], [3, -7]])  # desired turbine layouts
boundary = [(0, 0), (0, -10), (6, -10), (6, 0)]  # turbine boundaries


@pytest.fixture
def turbineXYZOptimizationProblem_generator():
    def _topfarm_obj(gradients, cost_comp=None, **kwargs):

        return TopFarmProblem(
            {'x': initial[:, 0], 'y': initial[:, 1]},
            cost_comp=cost_comp or CostModelComponent(['x', 'y'], 4, cost, gradients),
            constraints=[SpacingConstraint(2), XYBoundaryConstraint(boundary)],
            driver=EasyScipyOptimizeDriver(),
            **kwargs)
    return _topfarm_obj


def cost(x, y):
    return np.sum((x - optimal[:, 0])**2 + (y - optimal[:, 1])**2)


def income(x, y):
    return -np.sum((x - optimal[:, 0])**2 + (y - optimal[:, 1])**2)


def income_gradients(x, y):
    return (-(2 * x - 2 * optimal[:, 0]),
            -(2 * y - 2 * optimal[:, 1]))


def gradients(x, y):
    return ((2 * x - 2 * optimal[:, 0]),
            (2 * y - 2 * optimal[:, 1]))


def wrong_gradients(x, y):
    return ((2 * x - 2 * optimal[:, 0]) + 1,
            (2 * y - 2 * optimal[:, 1]))


def testTopFarmProblem_check_gradients(turbineXYZOptimizationProblem_generator):
    # Check that gradients check does not raise exception for correct gradients
    tf = turbineXYZOptimizationProblem_generator(gradients)
    tf.check_gradients(True)

    # Check that gradients check raises an exception for incorrect gradients
    tf = turbineXYZOptimizationProblem_generator(wrong_gradients)
    with pytest.raises(Warning, match="Mismatch between finite difference derivative of 'cost' wrt. 'x' and derivative computed in 'cost_comp' is"):
        tf.check_gradients()


def testTopFarmProblem_check_gradients_Income(turbineXYZOptimizationProblem_generator):
    # Check that gradients check does not raise exception for correct gradients
    cost_comp = CostModelComponent('xy', 4, income, income_gradients, income_model=True)
    tf = turbineXYZOptimizationProblem_generator(None, cost_comp)
    tf.check_gradients(True)

    # Check that gradients check raises an exception for incorrect gradients
    cost_comp = CostModelComponent('xy', 4, income, wrong_gradients, income_model=True)
    tf = turbineXYZOptimizationProblem_generator(None, cost_comp)
    with pytest.raises(Warning, match="Mismatch between finite difference derivative of 'cost' wrt. 'y' and derivative computed in 'cost_comp' is"):
        tf.check_gradients()


def testTopFarmProblem_evaluate_gradients(turbineXYZOptimizationProblem_generator):
    tf = turbineXYZOptimizationProblem_generator(gradients)
    np.testing.assert_array_equal(tf.evaluate_gradients()['aggr_cost']['x'], [[-6., -14., -8., -6.]])


def testTopFarmProblem_as_component(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    c = tf.as_component()
    npt.assert_equal(c.__class__, ProblemComponent)
    assert c.problem == tf


def testTopFarmProblem_get_DOE_list(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    npt.assert_array_equal(len(tf.get_DOE_list()), 27)
    (k, v), = tf.get_DOE_list()[1]
    assert k == "indeps.type"
    npt.assert_array_equal(v, [1, 0, 0])

    # npt.assert_array_equal(tf.get_DOE_list()[1], [[('indeps.turbineType', array([0., 0., 0.]))], [('indeps.turbineType', array([1., 0., 0.]))]])


def testTopFarmProblem_get_DOE_array(turbineTypeOptimizationProblem):
    tf = turbineTypeOptimizationProblem
    npt.assert_array_equal(tf.get_DOE_array().shape, (27, 1, 3))
    npt.assert_array_equal(tf.get_DOE_array()[:5], [[[0, 0, 0]],
                                                    [[1, 0, 0]],
                                                    [[2, 0, 0]],
                                                    [[0, 1, 0]],
                                                    [[1, 1, 0]]])


def testTopFarmProblem_turbine_positions():
    tf = xy3tb.get_tf()
    np.testing.assert_array_equal(tf.turbine_positions, xy3tb.initial)


def test_smart_start():
    xs_ref = [1.6, 1.6, 3.7]
    ys_ref = [1.6, 3.7, 1.6]

    x = np.arange(0, 5, 0.1)
    y = np.arange(0, 5, 0.1)
    YY, XX = np.meshgrid(y, x)
    ZZ = np.sin(XX) + np.sin(YY)
    min_spacing = 2.1
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(min_spacing)])
    np.random.seed(0)
    tf.smart_start(XX, YY, ZZ)
    npt.assert_array_almost_equal(tf.turbine_positions, np.array([xs_ref, ys_ref]).T)
    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, ZZ, 100)
        for x, y in tf.turbine_positions:
            circle = plt.Circle((x, y), min_spacing / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(x, y, 'rx')
        plt.axis('equal')
        plt.show()


def test_smart_start_boundary():
    xs_ref = [1.6, 1.6, 3.6]
    ys_ref = [1.6, 3.7, 2.3]

    x = np.arange(0, 5.1, 0.1)
    y = np.arange(0, 5.1, 0.1)
    YY, XX = np.meshgrid(y, x)
    ZZ = np.sin(XX) + np.sin(YY)
    min_spacing = 2.1
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(min_spacing),
                                   XYBoundaryConstraint([(0, 0), (5, 3), (5, 5), (0, 5)])])
    tf.smart_start(XX, YY, ZZ)
    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, ZZ, 100)
        plt.plot(tf.xy_boundary[:, 0], tf.xy_boundary[:, 1], 'k')
        for x, y in tf.turbine_positions:
            circle = plt.Circle((x, y), min_spacing / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(x, y, 'rx')

        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal(tf.turbine_positions, np.array([xs_ref, ys_ref]).T)


def test_smart_start_polygon_boundary():
    xs_ref = [1.6, 1.6, 3.6]
    ys_ref = [1.6, 3.7, 2.3]

    x = np.arange(0, 5.1, 0.1)
    y = np.arange(0, 5.1, 0.1)
    YY, XX = np.meshgrid(y, x)
    ZZ = np.sin(XX) + np.sin(YY)
    min_spacing = 2.1
    tf = xy3tb.get_tf(constraints=[SpacingConstraint(min_spacing),
                                   XYBoundaryConstraint([(0, 0), (5, 3), (5, 5), (0, 5)], 'polygon')])
    tf.smart_start(XX, YY, ZZ)
    if 0:
        import matplotlib.pyplot as plt
        plt.contourf(XX, YY, ZZ, 100)
        plt.plot(tf.xy_boundary[:, 0], tf.xy_boundary[:, 1], 'k')
        for x, y in tf.turbine_positions:
            circle = plt.Circle((x, y), min_spacing / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(x, y, 'rx')

        plt.axis('equal')
        plt.show()
    npt.assert_array_almost_equal(tf.turbine_positions, np.array([xs_ref, ys_ref]).T)



'''
Created on 20. apr. 2018

@author: mmpe
'''
import pytest


import numpy as np


class FusedWakeModel(object):

    def __init__(self, yml, version=None, **kwargs):
        """Description

        Parameters
        ----------
        yml: str, optional
            A WindIO `yml` file containing the description of the farm
        """
        from fusedwake.WindFarm import WindFarm
        self.windFarm = WindFarm(yml=yml)
        self.version = version or self.version
        self.wake_model = self.wake_model_cls(WF=self.windFarm, version=self.version, **kwargs)

    def __call__(self, turbine_positions, no_wake_wdir, no_wake_wsp, no_wake_ti):
        self.wake_model.update_position(turbine_positions)
        WD, WS, TI = (np.atleast_2d(v) for v in [no_wake_wdir, no_wake_wsp, no_wake_ti])
        assert WD.shape == WS.shape == TI.shape, "Shape of no_wake_wdir, no_wake_wsp and no_wake_ti must equal: %s != %s != %s" % (
            WD.shape, WS.shape, TI.shape)
        if len(WD.shape) == 3:
            WD, WS, TI = [np.mean(v, 0) for v in [WD, WS, TI]]
        self.run_wake_model(WS, WD, TI)
        p = self.wake_model.p_wt
        p = p.reshape(WD.shape + (self.windFarm.nWT,))
        return p.sum(2)  # sum over all turbines

    def run_wake_model(self, WS, WD, TI):
        self.wake_model(WS=WS.flatten(), WD=WD.flatten(), TI=TI.flatten())


class FusedWakeGCLWakeModel(FusedWakeModel):
    from fusedwake.gcl.interface import GCL
    wake_model_cls = GCL
    version = 'fort_gcl'


class FusedWakeNOJWakeModel(FusedWakeModel):
    from fusedwake.noj.interface import NOJ
    wake_model_cls = NOJ
    version = 'fort_noj'


def main():
    if __name__ == '__main__':
        try:
            import fusedwake
        except ModuleNotFoundError:
            return
        import os
        import matplotlib.pyplot as plt
        hornsrev_yml = os.path.dirname(fusedwake.__file__) + "/../examples/hornsrev.yml"
        noj, _ = wake_models = [FusedWakeNOJWakeModel(hornsrev_yml, K=.1), FusedWakeGCLWakeModel(hornsrev_yml)]
        tb_pos = noj.windFarm.pos
        print(noj(tb_pos.T, no_wake_wdir=270, no_wake_wsp=8, no_wake_ti=0.1))

        for wm, c in zip(wake_models, 'rk'):
            WS_cases = np.arange(11, 12)
            WD_cases = np.arange(0, 360, 1)
            WD_ms, WS_ms = np.meshgrid(WD_cases, WS_cases)
            p = wm(tb_pos.T, WD_ms, WS_ms, np.zeros_like(WS_ms) + .1)
            plt.xlabel("Wdir")
            plt.ylabel("Power")
            plt.plot(p.T, color=c, label=wm.wake_model_cls.__name__)
            plt.legend()
            print(p.sum())
        plt.show()


main()



import numpy as np
from topfarm.constraint_components import Constraint, ConstraintComponent
import topfarm


class SpacingConstraint(Constraint):
    def __init__(self, min_spacing, units=None):
        """Initialize SpacingConstraint

        Parameters
        ----------
        min_spacing : int or float
            Minimum spacing between turbines [m]
        """
        self.min_spacing = min_spacing
        self.const_id = 'spacing_comp_{}'.format(int(min_spacing))
        self.units = units

    @property
    def constraintComponent(self):
        return self.spacing_comp

    def _setup(self, problem):
        self.n_wt = problem.n_wt
        self.spacing_comp = SpacingComp(self.n_wt, self.min_spacing, self.const_id, self.units)
        problem.model.pre_constraints.add_subsystem(self.const_id, self.spacing_comp,
                                                    promotes=[topfarm.x_key, topfarm.y_key, 'penalty_' + self.const_id, 'wtSeparationSquared'])

    def setup_as_constraint(self, problem):
        self._setup(problem)
        zero = np.zeros(int(((self.n_wt - 1.) * self.n_wt / 2.)))
        problem.model.add_constraint('wtSeparationSquared', lower=zero + (self.min_spacing)**2)

    def setup_as_penalty(self, problem):
        self._setup(problem)


class SpacingComp(ConstraintComponent):
    """
    Calculates inter-turbine spacing for all turbine pairs.

    """

    def __init__(self, n_wt, min_spacing, const_id=None, units=None):
        super().__init__()
        self.n_wt = n_wt
        self.min_spacing = min_spacing
        self.const_id = const_id
        self.veclen = int((n_wt - 1.) * n_wt / 2.)
        self.units = units

    def setup(self):
        # Explicitly size input arrays
        self.add_input(topfarm.x_key, val=np.zeros(self.n_wt),
                       desc='x coordinates of turbines in wind dir. ref. frame', units=self.units)
        self.add_input(topfarm.y_key, val=np.zeros(self.n_wt),
                       desc='y coordinates of turbines in wind dir. ref. frame', units=self.units)
        self.add_output('penalty_' + self.const_id, val=0.0)
        # Explicitly size output array
        self.add_output('wtSeparationSquared', val=np.zeros(self.veclen),
                        desc='spacing of all turbines in the wind farm')
        # Sparse partial declaration
        cols = np.array([(i, j) for i in range(self.n_wt - 1)
                         for j in range(i + 1, self.n_wt)]).flatten()
        rows = np.repeat(np.arange(self.veclen), 2)

        self.declare_partials('wtSeparationSquared',
                              [topfarm.x_key, topfarm.y_key],
                              rows=rows, cols=cols)

    def compute(self, inputs, outputs):
        self.x = inputs[topfarm.x_key]
        self.y = inputs[topfarm.y_key]
        separation_squared = self._compute(self.x, self.y)
        outputs['wtSeparationSquared'] = separation_squared
        outputs['penalty_' + self.const_id] = -np.minimum(separation_squared - self.min_spacing**2, 0).sum()

    def _compute(self, x, y):
        n_wt = self.n_wt

        # compute distance matrixes
        dX, dY = [np.subtract(*np.meshgrid(xy, xy, indexing='ij')).T
                  for xy in [x, y]]
        dXY2 = dX**2 + dY**2
        # return upper triangle (above diagonal)
        return dXY2[np.triu_indices(n_wt, 1)]

    def compute_partials(self, inputs, J):
        # obtain necessary inputs
        x = inputs[topfarm.x_key]
        y = inputs[topfarm.y_key]

        dSdx, dSdy = self._compute_partials(x, y)
        # populate Jacobian dict
        J['wtSeparationSquared', topfarm.x_key] = dSdx.flatten()
        J['wtSeparationSquared', topfarm.y_key] = dSdy.flatten()

    def _compute_partials(self, x, y):
        # get number of turbines
        n_wt = self.n_wt

        # S = (xi-xj)^2 + (yi-yj)^2 = dx^2 + dy^2
        # dS/dx = [dS/dx_i, dS/dx_j]  = -2dx, 2dx

        # compute distance matrixes
        dX, dY = [np.subtract(*np.meshgrid(xy, xy, indexing='ij')).T
                  for xy in [x, y]]
        # upper triangle -> 1 row per WT pair [(0,1), (0,2),..(n-1,n)]
        dx, dy = dX[np.triu_indices(n_wt, 1)], dY[np.triu_indices(n_wt, 1)]

        dSdx = np.array([-2 * dx, 2 * dx]).T
        dSdy = np.array([-2 * dy, 2 * dy]).T
        return dSdx, dSdy

    def plot(self, ax=None):
        from matplotlib.pyplot import Circle
        import matplotlib.pyplot as plt
        ax = ax or plt.gca()

        def get_xy(xy):
            if not hasattr(self, xy):
                setattr(self, xy, dict(self.list_inputs(out_stream=None))[f'pre_constraints.{self.name}.{xy}']['value'])
            xy = getattr(self, xy)
            return xy if not isinstance(xy, tuple) else xy[0]

        for x, y in zip(get_xy('x'), get_xy('y')):
            circle = Circle((x, y), self.min_spacing / 2, color='k', ls='--', fill=False)
            ax.add_artist(circle)

    def satisfy(self, state, n_iter=100, step_size=0.1):
        x, y = [state[xy].astype(np.float) for xy in [topfarm.x_key, topfarm.y_key]]
        pair_i, pair_j = np.triu_indices(len(x), 1)
        for _ in range(n_iter):
            dist = self._compute(x, y)
            dx, dy = self._compute_partials(x, y)
            index = np.argmin(dist)

            if dist[index] < self.min_spacing**2:
                i, j = pair_i[index], pair_j[index]
                x[i] += dx[index, 0] * step_size
                x[j] += dx[index, 1] * step_size
                y[i] += dy[index, 0] * step_size
                y[j] += dy[index, 1] * step_size
            else:
                break
        state.update({topfarm.x_key: x, topfarm.y_key: y})
        return state



import topfarm
import openmdao.api as om
from topfarm.cost_models.utils.spanning_tree import spanning_tree
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
from topfarm.plotting import XYPlotComp, NoPlot

import numpy as np
import matplotlib.pylab as plt


class ElNetComp(om.ExplicitComponent):
    def initialize(self):
        self.options.declare('n_wt', types=int)

    def setup(self):
        self.add_input(topfarm.x_key, np.zeros(self.options['n_wt']), units='m')
        self.add_input(topfarm.y_key, np.zeros(self.options['n_wt']), units='m')


class ElNetLength(ElNetComp):
    def setup(self):
        super().setup()
        self.add_output('elnet_length', 0.0, units='m')

        # Note that this component is NOT derivative friendly, as the
        # electrical layout will jump from one turbine to another in a non-
        # continuous manner if you move the turbines
        self.declare_partials('elnet_length', [topfarm.x_key, topfarm.y_key],
                              method='fd')

    def compute(self, inputs, outputs):
        x, y = inputs[topfarm.x_key], inputs[topfarm.y_key]
        elnet_layout = spanning_tree(x, y)
        outputs['elnet_length'] = sum(list(elnet_layout.values()))


class PlotElNet(ElNetComp):
    def compute(self, inputs, outputs):
        x, y = inputs[topfarm.x_key], inputs[topfarm.y_key]
        elnet_layout = spanning_tree(x, y)
        indices = np.array(list(elnet_layout.keys())).T
        plt.plot(x[indices], y[indices], color='r')
        plt.plot(x, y, 'o')
        plt.axis('equal')
        plt.title(f'Total Length = {sum(list(elnet_layout.values()))}')


class ElNetCost(CostModelComponent):
    def __init__(self, n_wt, length_key='elnet_length', **kwargs):
        self.n_wt = n_wt
        self.length_key = length_key
        CostModelComponent.__init__(self, [(self.length_key, 0.0)], self.n_wt,
                                    self.cost, self.grad, objective=False,
                                    **kwargs)

    def initialize(self):
        self.options.declare('cost_per_meter')

    def cost(self, **kwargs):
        return self.options['cost_per_meter'] * kwargs[self.length_key]

    def grad(self, **kwargs):
        return [self.options['cost_per_meter']]


class XYElPlotComp(XYPlotComp):
    """Plotting component for turbine locations"""
    def plot_current_position(self, x, y):
        elnet_layout = spanning_tree(x, y)
        indices = np.array(list(elnet_layout.keys())).T
        plt.plot(x[indices], y[indices], color='r')
        for c, x_, y_ in zip(self.colors, x, y):
            self.ax.plot(x_, y_, 'o', color=c, ms=5)
            self.ax.plot(x_, y_, 'xk', ms=4)
        self.ax.plot([], [], 'xk', label='Current position')



import os
import pytest
import numpy as np
from topfarm.cost_models.cost_model_wrappers import AEPCostModelComponent
try:
    from py_colonel import PyColonel
    from py_colonel.py_colonel_exe import colonel_path
    fuga_path = colonel_path
except ImportError:
    PyColonel = object


class PyFuga(PyColonel):
    interface_version = 3

    def get_TopFarm_cost_component(self):
        n_wt = self.get_no_turbines()
        return AEPCostModelComponent(['x', 'y'], n_wt,
                                     lambda x, y, **kwargs: self.get_aep(np.array([x, y]).T)[0],  # only aep
                                     lambda x, y, **kwargs: self.get_aep_gradients(np.array([x, y]).T)[:2])  # only dAEPdx and dAEPdy


def main():
    if __name__ == '__main__':
        try:
            import py_colonel
        except ModuleNotFoundError:
            pytest.xfail("Colonel submodule not found\n")
        from py_colonel.py_colonel_lib import fugalib_path
        if os.path.isfile(fugalib_path) is False:
            pytest.xfail("Fugalib '%s' not found\n" % fugalib_path)
        pyFuga = PyFuga()
        pyFuga.setup(farm_name='Horns Rev 1',
                     turbine_model_path=fuga_path + 'LUTs-T/', turbine_model_name='Vestas_V80_(2_MW_offshore)[h=70.00]',
                     tb_x=[423974, 424033], tb_y=[6151447, 6150889],
                     mast_position=(0, 0, 70), z0=0.03, zi=400, zeta0=0,
                     farms_dir=fuga_path + 'LUTs-T/Farms/', wind_atlas_path=r'MyFarm\DEN05JBgr_7.813E_55.489N_7.4_5.lib')

        print(pyFuga.get_no_turbines())
        print(pyFuga.get_aep(np.array([[0, 0], [0, 1000]])))
        print(pyFuga.get_aep(np.array([[0, 1000], [0, 0]])))
        print(pyFuga.get_aep_gradients(np.array([[0, 0], [0, 100]])))


main()




from openmdao.drivers.doe_generators import FullFactorialGenerator
from openmdao.drivers.doe_driver import DOEDriver
from topfarm.cost_models.dummy import DummyCost
from topfarm.tests import npt

import numpy as np
from topfarm._topfarm import TopFarmProblem
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm.easy_drivers import EasyScipyOptimizeDriver


def test_turbineType_optimization():
    optimal = np.array([[1], [0]])
    tf = TopFarmProblem(
        design_vars={'type': (optimal[:, 0], 0, 1)},
        cost_comp=DummyCost(optimal_state=optimal, inputs=['type']),
        driver=DOEDriver(FullFactorialGenerator(2)))
    cost, state, _ = tf.optimize()
    assert cost == 0
    npt.assert_array_equal(state['type'], [1, 0])


def test_turbineXYZ_optimization():
    optimal = np.array([(5, 4, 3),
                        (3, 2, 1)])
    turbineXYZ = np.array([[0, 0, 0],
                           [2, 2, 2]])
    design_vars = {k: v for k, v in zip('xy', turbineXYZ.T)}
    design_vars['z'] = (turbineXYZ[:, 2], 1, 4)

    xy_boundary = [(0, 0), (5, 5)]
    tf = TopFarmProblem(
        design_vars=design_vars,
        cost_comp=DummyCost(optimal, 'xyz'),
        driver=EasyScipyOptimizeDriver(disp=False),
        constraints=[XYBoundaryConstraint(xy_boundary, 'square')])

    cost, state = tf.evaluate()
    assert cost == 52
    np.testing.assert_array_equal(state['x'], [0, 2])

    cost = tf.optimize()[0]
    assert cost < 1e6
    np.testing.assert_array_almost_equal(tf.turbine_positions, optimal[:, :2], 3)



"""
turbine_costsse_2017.py

Created by Arvydas Berzonskis  Goldwind 2017 based on turbine_costsse_2015.py 2015.
Copyright (c) NREL. All rights reserved.
"""
import numpy as np
import math
import matplotlib.pyplot as plt
from openmdao.core.explicitcomponent import ExplicitComponent
plt.close("all")

# Rotor
# -------------------------------------------------------------------------------


class economic_evaluation():

    def __init__(self, D_rotor_array, Power_rated_array, hub_height_array, aep_array, electrical_connection_cost=None, tower_cost=None):
        self.D_rotor_array = D_rotor_array
        self.Power_rated_array = Power_rated_array
        self.hub_height_array = hub_height_array
        self.aep_array = aep_array
        self.electrical_connection_cost = electrical_connection_cost
        self.tower_cost = tower_cost

    def calculate_irr(self):

        #        blade_type_array = [''] * len(D_rotor_array)
        # turbine_type_array = list(map(lambda x, y: self.reflection_turbine_type(x, y),
        #                           D_rotor_array, Power_rated_array))
        # weight_array = list(map(lambda x: self.simple_weight_height_trans(x),
        #                     hub_height_array))
        self.IRR = 0
        aep_vector = np.array([[float(item)] for item in self.aep_array])
        if sum(aep_vector) > 0:
            # turbine_type_vector = [[item] for item in turbine_type_array]
            machine_rating = np.array([[int(item)] for item in self.Power_rated_array])
            rotor_diameter = np.array([[int(item)] for item in self.D_rotor_array])
            hub_height = np.array([[int(item)] for item in self.hub_height_array])
            # weight_vector = [[float(item)] for item in weight_array]
            aep_vector = np.array([[float(item)] for item in self.aep_array])

        # def __init__(self, rotor_diameter, machine_rating, hub_height, aep_vector):
            # calculate the blade mass and cost
            # Baseline mode
            self.blade_B_mass = 3 * 0.1452 * (rotor_diameter / 2)**2.9158  # all 3 blades
            # Advanced mode
            self.blade_A_mass = 0.4948 * (rotor_diameter / 2)**2.53  # all 3 blades
            # Blade material cost escalator
            self.BCE = 1
            # Labor cost escalator
            self.GDPE = 1
            # Costs
            self.blade_B_costs = 3 * ((0.4019 * (rotor_diameter / 2)**3 - 955.24) * self.BCE +
                                      2.7445 * (rotor_diameter / 2)**2.5025 * self.GDPE) / (1 - 0.28)
            self.blade_A_costs = 3 * ((0.4019 * (rotor_diameter / 2)**3 - 21051) * self.BCE +
                                      2.7445 * (rotor_diameter / 2)**2.5025 * self.GDPE) / (1 - 0.28)

            # calculate the Hub cost and weight
            self.hub_mass = 0.954 * (self.blade_B_mass / 3) + 5680.3
            self.hub_cost = self.hub_mass * 4.25

            # Pitch mechanisms and bearings
            self.pich_bearing_mass = 0.1295 * self.blade_B_mass + 491.31
            self.pitch_system_mass = self.pich_bearing_mass * 1.328 + 555
            # Total pitch costs
            self.pitch_system_cost = 2.28 * (0.2106 * rotor_diameter**2.6578)  # All 3 blades

            # Spinner, nose cone
            self.nose_cone_mass = 18.5 * rotor_diameter - 520.5
            self.nose_cone_cost = self.nose_cone_mass * 5.57

            # Low-speed shaft
            ''' Notes might not be used for direct drive turbine costs'''
            self.low_speed_shaft_mass = 0.0142 * rotor_diameter**2.888
            self.low_speed_shaft_cost = 0.01 * rotor_diameter**2.887

            # Main bearings
            self.bearing_mass = (rotor_diameter * 8 / 600 - 0.033) * 0.0092 * rotor_diameter**2.5
            self.bearing_cost = 2 * self.bearing_mass * 17.6

            # Mecahnical brake, high-speed coupling and associated components
            self.brake_and_coupling_cost = 1.9894 * machine_rating - 0.1141
            self.brake_and_coupling_mass = self.brake_and_coupling_cost / 10.

            # Direct drive Generator
            # self.generator_mass=661.25*self.low_speed_shaft_torque**0.606
            self.generator_cost = machine_rating * 219.33

            # Variable-speed electronics
            self.variablespeed_electronics = machine_rating * 79.

            # Yaw Drive and Bearing
            self.yaw_system_mass = 1.6 * (0.00098 * rotor_diameter**3.314)
            self.yaw_system_cost = 2 * (0.0339 * rotor_diameter**2.964)

            # Mainframe - Direct Drive
            self.mainframe_mass = 1.228 * rotor_diameter**1.953
            self.mainframe_cost = 627.28 * rotor_diameter**0.85

            # Platforms and railings
            self.platform_railing_mass = 0.125 * self.mainframe_mass
            self.platform_railing_cost = self.platform_railing_mass * 8.7

            # Electrical connections
            if not self.electrical_connection_cost:
                self.electrical_connection_cost = machine_rating * 40.

            # Hydraulic and Cooling Systems
            self.hydraulic_cooling_system_mass = 0.08 * machine_rating
            self.hydraulic_cooling_system_cost = machine_rating * 12

            # Nacelle Cover
            self.nacelle_cost = 11.537 * machine_rating + 3849.7
            self.nacelle_mass = self.nacelle_cost / 10.

            # Control, Safety Sytem, Condition Monitoring
            self.control_cost = 35000.0

            # Tower
            # Baseline model
            if not self.tower_cost:
                self.tower_B_mass = 0.3973 * (math.pi * (rotor_diameter / 2)**2) * hub_height - 1414
                # Advanced model
                self.tower_A_mass = 0.2694 * (math.pi * (rotor_diameter / 2)**2) * hub_height - 1779
                self.tower_cost = self.tower_B_mass * 1.5

            # Foundation cost
            self.foundation_cost = 303.24 * (hub_height * (math.pi * (rotor_diameter / 2))**2)**0.4037

            # Transportation cost
            self.trasport_coeff = 1.581e-5 * machine_rating**2 - 0.0375 * machine_rating + 54.7
            self.trasport_cost = machine_rating * self.trasport_coeff

            # Roads and Civil work
            self.roads_civil_cost_fact = 2.17e-6 * machine_rating**2 - 0.0145 * machine_rating + 69.54
            self.roads_civil_cost_cost = self.roads_civil_cost_fact * machine_rating

            # Assembly and Installation
            self.assembly_and_installation_cost = 1.965 * (hub_height * rotor_diameter)**1.1736

            # Electrical Interface and connections
            self.electrical_interface_fact = 3.49e-6 * machine_rating**2 - 0.0221 * machine_rating + 109.7
            self.electrical_interface_cost = machine_rating * self.electrical_interface_fact

            # Annual operating expenses - AOE
            self.LCC = 0.00108 * sum(aep_vector)
            self.OandM = 0.007 * sum(aep_vector)
            self.LRC = sum(10.7 * machine_rating)
            self.AOE = self.LCC + (self.OandM + self.LRC) / sum(aep_vector)

            # I do not know how to calculate the weight
            # self.variable_speed_elec_mass_cost=18.8*variable_speed_elec_mass

            self.cost = 1.33 * (self.blade_B_costs + self.hub_cost + self.pitch_system_cost + self.nose_cone_cost + self.bearing_cost +
                                self.brake_and_coupling_cost + self.generator_cost + self.variablespeed_electronics + self.yaw_system_cost +
                                self.mainframe_cost + self.platform_railing_cost + self.electrical_connection_cost +
                                self.hydraulic_cooling_system_cost + self.nacelle_cost + self.control_cost + self.tower_cost +
                                self.foundation_cost + self.trasport_cost + self.roads_civil_cost_cost + self.assembly_and_installation_cost +
                                self.electrical_interface_cost)  # the ratio of 1.33 is to account for inflation since 2003.

            self.cost[self.aep_array == 0] = 0

            self.CWF = []

            for i in range(1, 20):
                if i == 1:
                    self.CWF.append(int(0.1 * sum(aep_vector) - sum(self.cost) - self.AOE))
                else:
                    self.CWF.append(int(0.1 * sum(aep_vector) - self.AOE))

            self.IRR = 100 * np.irr(self.CWF)
        return self.IRR


def main():
    if __name__ == '__main__':
        Drotor_vector = np.array([121.0, 115.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0,
                                  121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 121.0, 115.0])
        power_rated_vector = np.array([2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000.,
                                       2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000., 2000.])
        hub_height_vector = np.array([105., 105., 105., 105., 105., 105., 105., 105., 105., 105., 105., 105., 105., 105.,
                                      105., 105., 105., 105., 105., 105., 105., 105., 105., 105.])
        aep_vector = np.array([7719409.76158781, 7703877.645459095, 6034197.878645113, 6494510.809393061, 8147512.144511054,
                               7369147.974147366, 8681084.747836143, 7147269.951808259, 6141634.901501132, 5682294.113683184,
                               6859259.383656042, 7447180.72443833, 5761350.543959712, 6755014.03150229, 6337664.079841179,
                               5151511.151202235, 5456811.610073241, 6722954.6423807535, 7185180.112344236, 7215858.2123201955,
                               7136356.084546783, 6267802.418413456, 5899094.2746176915, 6592877.854876981])
        # initialize class
        Turbine = economic_evaluation(Drotor_vector, power_rated_vector, hub_height_vector, aep_vector)  # Rotor diameter; Power rated; Hub Height
        # run the method for IRR calculations
        Turbine.calculate_irr()

        print('Wind turbine configuration costs', Turbine.cost)
        print('IRR', Turbine.IRR)

    #    fig,ax = plt.subplots()
    #    ax.plot(Drotor_vector,economic_evaluation.cost)
    #    ax.set(xlabel='Rotor diameter [m]', ylabel='Initial costs [$]')

        # just the cost bar lot
        barplotvector = np.array([Turbine.blade_B_costs[0], Turbine.hub_cost[0], Turbine.pitch_system_cost[0], Turbine.nose_cone_cost[0],
                                  Turbine.bearing_cost[0], Turbine.brake_and_coupling_cost[0], Turbine.generator_cost[0],
                                  Turbine.variablespeed_electronics[0], Turbine.yaw_system_cost[0], Turbine.mainframe_cost[0],
                                  Turbine.platform_railing_cost[0], Turbine.electrical_connection_cost[0],
                                  Turbine.hydraulic_cooling_system_cost[0], Turbine.nacelle_cost[0], 35000,
                                  Turbine.tower_cost[0], Turbine.foundation_cost[0], Turbine.trasport_cost[0], Turbine.roads_civil_cost_cost[0],
                                  Turbine.assembly_and_installation_cost[0], Turbine.electrical_interface_cost[0]] / Turbine.cost[0])
        N = 21
        ind = np.arange(N)  # the x locations for the groups
        width = 0.5       # the width of the bars

        plt.rcParams.update({'figure.autolayout': True})
        fig, ax = plt.subplots()
        bar = ax.barh(ind, barplotvector, width, color='b')

        # add some text for labels, title and axes ticks
        ax.set_xlabel('Procentage of total costs [%]')
        ax.set_title('Wind turbine components costs')
        ax.set_yticks(ind + width / 2)
        ax.set_yticklabels(('Blades', 'Hub', 'Pitch System', 'Nose cone', 'Bearing', 'Break and coupling', 'Generator',
                            'Variable speed electornics', 'Yaw system', 'Mainframe', 'Platform railing', 'Electical connections',
                            'Hydraulic cooling systems', 'Nacelle cover', 'Control', 'Tower', 'Foundation', 'Transportation',
                            'Roads', 'Assembly and installation', 'Electrical interface'))
        labels = ax.get_xticklabels()
        plt.setp(labels, rotation=45, horizontalalignment='right')
        plt.show()


main()



from abc import ABC, abstractmethod
from openmdao.core.explicitcomponent import ExplicitComponent


class Constraint(ABC):

    @abstractmethod
    def setup_as_constraint(self):
        pass

    @abstractmethod
    def setup_as_penalty(self):
        pass


class ConstraintComponent(ExplicitComponent, ABC):
    def __init__(self, **kwargs):
        ExplicitComponent.__init__(self, **kwargs)

    @abstractmethod
    def satisfy(self, state):
        pass

    def plot(self, ax):
        pass



# -*- coding: utf-8 -*-
"""
Setup file for Topfarm2
"""


import os
from git_utils import write_vers
from setuptools import setup, find_packages

repo = os.path.dirname(__file__)
version = write_vers(vers_file='topfarm/__init__.py', repo=repo, skip_chars=1)

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name='topfarm', 
      version=version,  
      description='Topfarm - Wind farm optimization using OpenMDAO',
	  long_description=read('README'),
      url='https://gitlab.windenergy.dtu.dk/TOPFARM/topfarm2',
      author='DTU Wind Energy',  
      author_email='dave@dtu.dk',
      license='MIT',
      packages=find_packages(exclude=["*.examples", "*.examples.*", "examples.*", "examples"]),
      install_requires=[
        'matplotlib',  # for plotting
        'numpy',  # for numerical calculations
        'openmdao==2.6',  # for optimization
        'pytest',  # for testing
        'pytest-cov',  # for calculating coverage
		'py_wake>=1.0.11',  # for calculating AEP
        'scipy',  # constraints
        'sphinx',  # generating documentation
        'sphinx_rtd_theme',  # docs theme
        'scikit-learn',  # load surrogate
      ],
      zip_safe=True)



from openmdao.api import ExplicitComponent


class AggregatedCost(ExplicitComponent):
    '''
    Component that collects the cost and penalties and evaluates the aggregated cost
    '''
    def __init__(self, constraints_as_penalty, constraints, post_constraints):
        super().__init__()
        self.constraints_as_penalty = constraints_as_penalty
        self.constraints = constraints
        self.post_constraints = post_constraints

    def setup(self):
        self.add_input('cost', val=0)
        if self.constraints_as_penalty:
            if len(self.constraints) > 0:
                self.add_input('penalty', val=0)
            if len(self.post_constraints) > 0:
                self.add_input('post_penalty', val=0)
        self.add_output('aggr_cost', val=0)
        self.declare_partials(['*'], ['*'], method='fd')

    def compute(self, inputs, outputs):
        pen = 0
        if 'penalty' in inputs:
            pen += inputs['penalty']
        if 'post_penalty' in inputs:
            pen += inputs['post_penalty']
        if self.constraints_as_penalty and pen > 0:
            outputs['aggr_cost'] = pen + 10**10
        else:
            outputs['aggr_cost'] = inputs['cost']



import numpy as np


def smart_start(XX, YY, ZZ, N_WT, min_space, radius=None):
    """Selects the a number of gridpoints (N_WT) in the grid defined by x and y,
    where ZZ has the maximum value, while chosen points spacing (min_space)
    is respected.

    Parameters
    ----------
    XX : array_like
        x coordinates
    YY : array_like
        y coordinates
    ZZ : array_like
        Values at (XX, YY) of the desired variable in the
        grid points. This could be e.g. the AEP or wind speed.
    N_WT : integer
        number of wind turbines
    min_space: float
        minimum space between turbines

    Returns
    -------
    Positions where the aep or wsp is highest, while
    respecting the minimum spacing requirement.

    Notes
    -----
    XX, YY and ZZ can be 1D or 2D, but must have same size
    """
    ZZ_is_func = hasattr(ZZ, '__call__')
    if ZZ_is_func:
        arr = np.array([XX.flatten(), YY.flatten()])
    else:
        arr = np.array([XX.flatten(), YY.flatten(), ZZ.flatten()])
    if radius is not None and (np.diff(np.sort(np.unique(arr[0]))).min() > radius and
                               (np.diff(np.sort(np.unique(arr[1]))).min() > radius)):
        radius = None

    xs, ys = [], []
    for _ in range(N_WT):
        try:
            if ZZ_is_func:
                z = ZZ(arr[0], arr[1], xs, ys)
                ind = np.where(z == z.max())[0]
                if radius is None or len(ind) == len(z):
                    max_ind = ind[np.random.randint(len(ind))]
                else:
                    x, y = arr
                    z_rotor = np.array([np.mean(z[(x[i] - x)**2 + (y[i] - y)**2 < radius**2]) for i in ind])
                    ind = ind[np.where(z_rotor == z_rotor.max())[0]]
                    max_ind = ind[np.random.randint(len(ind))]
            else:
                z = arr[2]
                ind = np.where(z == z.max())[0]
                max_ind = ind[np.random.randint(len(ind))]

            x0 = arr[0][max_ind]
            y0 = arr[1][max_ind]
            xs.append(x0)
            ys.append(y0)
            index = np.where((arr[0] - x0)**2 + (arr[1] - y0)**2 >= min_space**2)[0]
            arr = arr[:, index]
            import os

        except ValueError:
            xs.append(np.nan)
            ys.append(np.nan)
            print('Could not respect the spacing constraint')
    return xs, ys


def main():
    if __name__ == '__main__':
        import matplotlib.pyplot as plt
        N_WT = 30
        min_space = 2.1

        x = np.arange(0, 20, 0.1)
        y = np.arange(0, 10, 0.1)
        YY, XX = np.meshgrid(y, x)
        val = np.sin(XX) + np.sin(YY)
        min_space = 2.1
        xs, ys = smart_start(XX, YY, val, N_WT, min_space)
        c = plt.contourf(XX, YY, val, 100)
        plt.colorbar(c)
        for i in range(N_WT):
            circle = plt.Circle((xs[i], ys[i]), min_space / 2, color='b', fill=False)
            plt.gcf().gca().add_artist(circle)
            plt.plot(xs[i], ys[i], 'rx')
        plt.axis('equal')
        plt.show()


main()



import os
import numpy as np
import multiprocessing
from topfarm.constraint_components.boundary_component import BoundaryComp
from topfarm.cost_models.dummy import DummyCost
from openmdao.drivers.doe_generators import UniformGenerator
from topfarm.constraint_components.boundary import XYBoundaryConstraint
from topfarm._topfarm import TopFarmProblem
import threading


class ParallelRunner():

    def __init__(self, processes=None):
        """
        Parameters
        ----------
        processes : int or None, optional
            Number of processes passed to multiprocessing.Pool
        """
        self.pool = multiprocessing.Pool(processes)

    def __call__(self, state_lst, seq_runner):
        """Run in parallel

        Parameters
        ----------
        state_lst : list
            List of states (as returned by TopFarmProblem.get_DOE_list()
        seq_runner : function
            function for sequential run. Interface must be:
            def f(lst):
                tf = TopfarmProblem(...)
                return tf.optimize()

        Returns
        -------
        best : (cost, state, recorder)
            best result
        results : [(cost1, state1, recorder1),...]
            all results
        """

        indexes = np.round(np.linspace(0, len(state_lst), self.pool._processes + 1)).astype(np.int)
        seq_lst = [state_lst[i1:i2] for i1, i2 in zip(indexes[:-1], indexes[1:])]

        results = self.pool.map(seq_runner, seq_lst)
        best = results[np.argmin([r[0] for r in results])]
        return best, results


# ===============================================================================
# Example functions
# ===============================================================================
def get_InitialXYZOptimizationProblem(driver):
    optimal = [(1, 0, 4),
               (0, 1, 3)]
    return TopFarmProblem(
        design_vars={'x': [0, 2], 'y': [0, 2], 'z': ([0, 2], 3, 4)},
        cost_comp=DummyCost(optimal, ['x', 'y', 'z']),
        constraints=[XYBoundaryConstraint([(10, 6), (11, 8)], 'rectangle')],
        driver=driver)


def seq_runner_example(lst):
    print("%d cases executed by thread: %s" % (len(lst), threading.get_ident()))
    return get_InitialXYZOptimizationProblem(lst).optimize()
# ===============================================================================
#
# ===============================================================================


def main():
    if __name__ == '__main__':

        lst = get_InitialXYZOptimizationProblem(driver=UniformGenerator(200)).get_DOE_list()
        print("Current thread: %s" % threading.get_ident())
        print("\nRun on two processors")
        # run in parallel
        par_runner = ParallelRunner(2)
        (cost, state, recorder), results = par_runner(lst, seq_runner_example)
        print("Minimum cost: %.2f" % cost)

        print("\nRun on one processor")
        # run sequential
        cost, state, recorder = seq_runner_example(lst)
        print("Minimum cost: %.2f" % cost)


main()



import matplotlib.pyplot as plt
import numpy as np
from topfarm.plotting import XYPlotComp
from topfarm.cost_models.cost_model_wrappers import CostModelComponent
import topfarm


class DummyCost(CostModelComponent):
    """Sum of squared error between current and optimal state
       Evaluates the equation
       f(x,..) = SUM((x_i - optx_i)^2 + ...).
    """

    def __init__(self, optimal_state, inputs=[topfarm.x_key, topfarm.y_key]):
        """
        Parameters
        ----------
        optimal_state : array_like, dim=(#wt,#inputs)
            optimal state array
        inputs : array_like
            list of input names
        """
        self.optimal_state = np.array(optimal_state)
        self.n_wt = self.optimal_state.shape[0]

        CostModelComponent.__init__(self, inputs, self.n_wt, self.cost, self.grad)

    def cost(self, **kwargs):
        opt = self.optimal_state
        return np.sum([(kwargs[n] - opt[:, i])**2 for i, n in enumerate(self.input_keys)])

    def grad(self, **kwargs):
        opt = self.optimal_state
        return [(2 * kwargs[n] - 2 * opt[:, i]) for i, n in enumerate(self.input_keys)]


class TurbineTypeDummyCost(DummyCost):
    def __init__(self, optimal):
        if len(np.array(optimal).shape) == 1:
            optimal = np.array([optimal]).T
        DummyCost.__init__(self, optimal, inputs=[topfarm.type_key])


class DummyCostPlotComp(XYPlotComp):
    def __init__(self, optimal, memory=10, delay=0.001, **kwargs):
        super().__init__(memory, delay, **kwargs)
        self.optimal = optimal

    def init_plot(self, boundary):
        opt_x, opt_y = np.array(self.optimal).T[0:2]
        XYPlotComp.init_plot(self, boundary)
        for c, optx, opty in zip(self.colors, opt_x, opt_y):
            plt.plot(optx, opty, 'ko', ms=6)
            plt.plot(optx, opty, 'o', color=c, ms=4)
        plt.plot([], [], 'o', color='k', markeredgecolor='k',
                 markerfacecolor='#00000000', ms=6, label='Optimal position')



import os
import numpy as np
from topfarm.cost_models.dummy import DummyCostPlotComp
from topfarm.tests import uta
import pytest
from topfarm.tests.test_files import xy3tb


def test_optimize_3tb():
    """Optimize 3-turbine layout and check final positions

        The farm boundaries and min spacing are chosen such that the desired
        turbine positions are not within the boundaries or constraints.
    """

    dec_prec = 4  # decimal precision for comparison

    tf = xy3tb.get_tf()
    tf.optimize()
    tb_pos = tf.turbine_positions

    tol = 1e-6
    uta.assertGreater(sum((tb_pos[2] - tb_pos[0])**2), 2**2 - tol)  # check min spacing
    uta.assertLess(tb_pos[1][0], 6 + tol)  # check within border
    np.testing.assert_array_almost_equal(tb_pos[:, :2], xy3tb.optimal, dec_prec)


def testDummyCostPlotComp():
    if os.name == 'posix' and "DISPLAY" not in os.environ:
        pytest.xfail("No display")

    tf = xy3tb.get_tf(plot_comp=DummyCostPlotComp(xy3tb.desired))
    tf.evaluate()
    tf.optimize()
